"""
Computes metrics for assessing quality of single units.

Run the following to set-up the workspace to run the docstring examples:
>>> import brainbox as bb
>>> import alf.io as aio
>>> import numpy as np
>>> import matplotlib.pyplot as plt
>>> import ibllib.ephys.spikes as e_spks
# (*Note, if there is no 'alf' directory, make 'alf' directory from 'ks2' output directory):
>>> e_spks.ks2_to_alf(path_to_ks_out, path_to_alf_out)
# Load the alf spikes bunch and clusters bunch, and get a units bunch.
>>> spks_b = aio.load_object(path_to_alf_out, 'spikes')
>>> clstrs_b = aio.load_object(path_to_alf_out, 'clusters')
>>> units_b = bb.processing.get_units_bunch(spks_b)  # may take a few mins to compute

TODO add spikemetrics as dependency?
TODO metrics that could be added: iso_dist, l_ratio, d_prime, nn_hit, nn_miss, sil
"""

import time

import numpy as np
import scipy.ndimage.filters as filters
import scipy.stats as stats

import brainbox as bb
from brainbox.core import Bunch
from brainbox.processing import bincount2D
from ibllib.io import spikeglx

# Parameters to be used in `quick_unit_metrics`
METRICS_PARAMS = {
    'presence_window': 10,
    'refractory_period': 0.0015,
    'min_isi': 0.0001,
    'spks_per_bin_for_missed_spks_est': 10,
    'std_smoothing_kernel_for_missed_spks_est': 4,
    'min_num_bins_for_missed_spks_est': 50,
}


def unit_stability(units_b, units=None, feat_names=['amps'], dist='norm', test='ks'):
    """
    Computes the probability that the empirical spike feature distribution(s), for specified
    feature(s), for all units, comes from a specific theoretical distribution, based on a specified
    statistical test. Also computes the coefficients of variation of the spike feature(s) for all
    units.

    Parameters
    ----------
    units_b : bunch
        A units bunch containing fields with spike information (e.g. cluster IDs, times, features,
        etc.) for all units.
    units : array-like (optional)
        A subset of all units for which to create the bar plot. (If `None`, all units are used)
    feat_names : list of strings (optional)
        A list of names of spike features that can be found in `spks` to specify which features to
        use for calculating unit stability.
    dist : string (optional)
        The type of hypothetical null distribution for which the empirical spike feature
        distributions are presumed to belong to.
    test : string (optional)
        The statistical test used to compute the probability that the empirical spike feature
        distributions come from `dist`.

    Returns
    -------
    p_vals_b : bunch
        A bunch with `feat_names` as keys, containing a ndarray with p-values (the probabilities
        that the empirical spike feature distribution for each unit comes from `dist` based on
        `test`) for each unit for all `feat_names`.
    cv_b : bunch
        A bunch with `feat_names` as keys, containing a ndarray with the coefficients of variation
        of each unit's empirical spike feature distribution for all features.

    See Also
    --------
    plot.feat_vars

    Examples
    --------
    1) Compute 1) the p-values obtained from running a one-sample ks test on the spike amplitudes
    for each unit, and 2) the variances of the empirical spike amplitudes distribution for each
    unit. Create a histogram of the variances of the spike amplitudes for each unit, color-coded by
    depth of channel of max amplitudes. Get cluster IDs of those units which have variances greater
    than 50.
        >>> p_vals_b, variances_b = bb.metrics.unit_stability(units_b)
        # Plot histograms of variances color-coded by depth of channel of max amplitudes
        >>> fig = bb.plot.feat_vars(units_b, feat_name='amps')
        # Get all unit IDs which have amps variance > 50
        >>> var_vals = np.array(tuple(variances_b['amps'].values()))
        >>> bad_units = np.where(var_vals > 50)
    """

    # Get units.
    if not (units is None):  # we're using a subset of all units
        unit_list = list(units_b[feat_names[0]].keys())
        # for each `feat` and unit in `unit_list`, remove unit from `units_b` if not in `units`
        for feat in feat_names:
            [units_b[feat].pop(unit) for unit in unit_list if not (int(unit) in units)]
    unit_list = list(units_b[feat_names[0]].keys())  # get new `unit_list` after removing units

    # Initialize `p_vals` and `variances`.
    p_vals_b = Bunch()
    cv_b = Bunch()

    # Set the test as a lambda function (in future, more tests can be added to this dict)
    tests = \
        {
            'ks': lambda x, y: stats.kstest(x, y)
        }
    test_fun = tests[test]

    # Compute the statistical tests and variances. For each feature, iteratively get each unit's
    # p-values and variances, and add them as keys to the respective bunches `p_vals_feat` and
    # `variances_feat`. After iterating through all units, add these bunches as keys to their
    # respective parent bunches, `p_vals` and `variances`.
    for feat in feat_names:
        p_vals_feat = Bunch((unit, 0) for unit in unit_list)
        cv_feat = Bunch((unit, 0) for unit in unit_list)
        for unit in unit_list:
            # If we're missing units/features, create a NaN placeholder and skip them:
            if len(units_b['times'][str(unit)]) == 0:
                p_val = np.nan
                cv = np.nan
            else:
                # compute p_val and var for current feature
                _, p_val = test_fun(units_b[feat][unit], dist)
                cv = np.var(units_b[feat][unit]) / np.mean(units_b[feat][unit])
            # Append current unit's values to list of units' values for current feature:
            p_vals_feat[str(unit)] = p_val
            cv_feat[str(unit)] = cv
        p_vals_b[feat] = p_vals_feat
        cv_b[feat] = cv_feat

    return p_vals_b, cv_b


def missed_spikes_est(feat, spks_per_bin=20, sigma=4, min_num_bins=50):
    """
    Computes the approximate fraction of spikes missing (i.e. a pseudo false negative estimate)
    from a spike feature distribution for a given unit, assuming the distribution is symmetric.

    Inspired by metric described in Hill et al. (2011) J Neurosci 31: 8699-8705.

    Parameters
    ----------
    feat : ndarray
        The spikes' feature values (e.g. amplitudes)
    spks_per_bin : int (optional)
        The number of spikes per bin from which to compute the spike feature histogram.
    sigma : int (optional)
        The standard deviation for the gaussian kernel used to compute the pdf from the spike
        feature histogram.
    min_num_bins : int (optional)
        The minimum number of bins used to compute the spike feature histogram.

    Returns
    -------
    fraction_missing : float
        The fraction of missing spikes (0-0.5). *Note: If more than 50% of spikes are missing, an
        accurate estimate isn't possible.
    pdf : ndarray
        The computed pdf of the spike feature histogram.
    cutoff_idx : int
        The index for `pdf` at which point `pdf` is no longer symmetrical around the peak. (This
        is returned for plotting purposes).

    See Also
    --------
    plot.missed_spikes_est

    Examples
    --------
    1) Determine the fraction of spikes missing from unit 1 based on the recorded unit's spike
    amplitudes, assuming the distribution of the unit's spike amplitudes is symmetric.
        # Get unit 1 amplitudes from a unit bunch, and compute fraction spikes missing.
        >>> feat = units_b['amps']['1']
        >>> fraction_missing = bb.plot.missed_spikes_est(feat)
    """

    # Ensure minimum number of spikes requirement is met.
    error_str = 'The number of spikes in this unit is {0}, ' \
                'but it must be at least {1}'.format(feat.size, spks_per_bin * min_num_bins)
    assert (feat.size > (spks_per_bin * min_num_bins)), error_str

    # compute the spike feature histogram and pdf:
    num_bins = np.int(feat.size / spks_per_bin)
    hist, bins = np.histogram(feat, num_bins, density=True)
    pdf = filters.gaussian_filter1d(hist, sigma)

    # Find where the distribution stops being symmetric around the peak:
    peak_idx = np.argmax(pdf)
    max_idx_sym_around_peak = np.argmin(np.abs(pdf[peak_idx:] - pdf[0]))
    cutoff_idx = peak_idx + max_idx_sym_around_peak

    # compute fraction missing from the tail of the pdf (the area where pdf stops being
    # symmetric around peak).
    fraction_missing = np.sum(pdf[cutoff_idx:]) / np.sum(pdf)
    fraction_missing = 0.5 if (fraction_missing > 0.5) else fraction_missing

    return fraction_missing, pdf, cutoff_idx


def wf_similarity(wf1, wf2):
    """
    Computes a unit normalized spatiotemporal similarity score between two sets of waveforms.
    This score is based on how waveform shape correlates for each pair of spikes between the
    two sets of waveforms across space and time. The shapes of the arrays of the two sets of
    waveforms must be equal.

    Parameters
    ----------
    wf1 : ndarray
        An array of shape (#spikes, #samples, #channels).
    wf2 : ndarray
        An array of shape (#spikes, #samples, #channels).

    Returns
    -------
    s: float
        The unit normalized spatiotemporal similarity score.

    See Also
    --------
    io.extract_waveforms
    plot.single_unit_wf_comp

    Examples
    --------
    1) Compute the similarity between the first and last 100 waveforms for unit1, across the 20
    channels around the channel of max amplitude.
        # Get the channels around the max amp channel for the unit, two sets of timestamps for the
        # unit, and the two corresponding sets of waveforms for those two sets of timestamps.
        # Then compute `s`.
        >>> max_ch = clstrs_b['channels'][1]
        >>> if max_ch < 10:  # take only channels greater than `max_ch`.
        >>>     ch = np.arange(max_ch, max_ch + 20)
        >>> elif (max_ch + 10) > 385:  # take only channels less than `max_ch`.
        >>>    ch = np.arange(max_ch - 20, max_ch)
        >>> else:  # take `n_c_ch` around `max_ch`.
        >>>     ch = np.arange(max_ch - 10, max_ch + 10)
        >>> ts1 = units_b['times']['1'][:100]
        >>> ts2 = units_b['times']['1'][-100:]
        >>> wf1 = bb.io.extract_waveforms(path_to_ephys_file, ts1, ch)
        >>> wf2 = bb.io.extract_waveforms(path_to_ephys_file, ts2, ch)
        >>> s = bb.metrics.wf_similarity(wf1, wf2)

    TODO check `s` calculation:
    take median of waveforms
    xcorr all waveforms with median, and divide by autocorr of all waveforms
    profile
    for two sets of units: xcorr(cl1, cl2) / (sqrt autocorr(cl1) * autocorr(cl2))
    """

    # Remove warning for dividing by 0 when calculating `s` (this is resolved by using
    # `np.nan_to_num`)
    import warnings
    warnings.filterwarnings('ignore', r'invalid value encountered in true_divide')
    assert wf1.shape == wf2.shape, ('The shapes of the sets of waveforms are inconsistent ({})'
                                    '({})'.format(wf1.shape, wf2.shape))

    # Get number of spikes, samples, and channels of waveforms.
    n_spks = wf1.shape[0]
    n_samples = wf1.shape[1]
    n_ch = wf1.shape[2]

    # Create a matrix that will hold the similarity values of each spike in `wf1` to `wf2`.
    # Iterate over both sets of spikes, computing `s` for each pair.
    similarity_matrix = np.zeros((n_spks, n_spks))
    for spk1 in range(n_spks):
        for spk2 in range(n_spks):
            s_spk = \
                np.sum(np.nan_to_num(
                    wf1[spk1, :, :] * wf2[spk2, :, :] /
                    np.sqrt(wf1[spk1, :, :] ** 2 * wf2[spk2, :, :] ** 2))) / (n_samples * n_ch)
            similarity_matrix[spk1, spk2] = s_spk

    # Return mean of similarity matrix
    s = np.mean(similarity_matrix)
    return s


def firing_rate_cv(ts, hist_win=0.01, fr_win=0.5, n_bins=10):
    """
    Computes the coefficient of variation of the firing rate: the ratio of the standard
    deviation to the mean.

    Parameters
    ----------
    ts : ndarray
        The spike timestamps from which to compute the firing rate.
    hist_win : float (optional)
        The time window (in s) to use for computing spike counts.
    fr_win : float (optional)
        The time window (in s) to use as a moving slider to compute the instantaneous firing rate.
    n_bins : int (optional)
        The number of bins in which to compute a coefficient of variation of the firing rate.

    Returns
    -------
    cv : float
        The mean coefficient of variation of the firing rate of the `n_bins` number of coefficients
        computed.
    cvs : ndarray
        The coefficients of variation of the firing for each bin of `n_bins`.
    fr : ndarray
        The instantaneous firing rate over time (in hz).

    See Also
    --------
    singlecell.firing_rate
    plot.firing_rate

    Examples
    --------
    1) Compute the coefficient of variation of the firing rate for unit 1 from the time of its
    first to last spike, and compute the coefficient of variation of the firing rate for unit 2
    from the first to second minute.
        >>> ts_1 = units_b['times']['1']
        >>> ts_2 = units_b['times']['2']
        >>> ts_2 = np.intersect1d(np.where(ts_2 > 60)[0], np.where(ts_2 < 120)[0])
        >>> cv, cvs, fr = bb.metrics.firing_rate_cv(ts_1)
        >>> cv_2, cvs_2, fr_2 = bb.metrics.firing_rate_cv(ts_2)
    """

    # Compute overall instantaneous firing rate and firing rate for each bin.
    fr = bb.singlecell.firing_rate(ts, hist_win=hist_win, fr_win=fr_win)
    bin_sz = np.int(fr.size / n_bins)
    fr_binned = np.array([fr[(b * bin_sz):(b * bin_sz + bin_sz)] for b in range(n_bins)])

    # Compute coefficient of variations of firing rate for each bin, and the mean c.v.
    cvs = np.std(fr_binned, axis=1) / np.mean(fr_binned, axis=1)
    # NaNs from zero spikes are turned into 0's
    # cvs[np.isnan(cvs)] = 0 nan's can happen if neuron doesn't spike in a bin
    cv = np.mean(cvs)

    return cv, cvs, fr


def firing_rate_fano_factor(ts, hist_win=0.01, fr_win=0.5, n_bins=10):
    '''
    Computes the fano factor of the firing rate: the ratio of the variance to the mean.
    (Almost identical to coeff. of variation)

    Parameters
    ----------
    ts : ndarray
        The spike timestamps from which to compute the firing rate.
    hist_win : float
        The time window (in s) to use for computing spike counts.
    fr_win : float
        The time window (in s) to use as a moving slider to compute the instantaneous firing rate.
    n_bins : int (optional)
        The number of bins in which to compute a fano factor of the firing rate.

    Returns
    -------
    ff : float
        The mean fano factor of the firing rate of the `n_bins` number of factors
        computed.
    ffs : ndarray
        The fano factors of the firing for each bin of `n_bins`.
    fr : ndarray
        The instantaneous firing rate over time (in hz).

    See Also
    --------
    singlecell.firing_rate
    plot.firing_rate

    Examples
    --------
    1) Compute the fano factor of the firing rate for unit 1 from the time of its
    first to last spike, and compute the fano factor of the firing rate for unit 2
    from the first to second minute.
        >>> ts_1 = units_b['times']['1']
        >>> ts_2 = units_b['times']['2']
        >>> ts_2 = np.intersect1d(np.where(ts_2 > 60)[0], np.where(ts_2 < 120)[0])
        >>> ff, ffs, fr = bb.metrics.firing_rate_fano_factor(ts_1)
        >>> ff_2, ffs_2, fr_2 = bb.metrics.firing_rate_fano_factor(ts_2)
    '''

    # Compute overall instantaneous firing rate and firing rate for each bin.
    fr = bb.singlecell.firing_rate(ts, hist_win=hist_win, fr_win=fr_win)
    # this procedure can cut off data at the end, up to n_bins last timesteps
    bin_sz = np.int(fr.size / n_bins)
    fr_binned = np.array([fr[(b * bin_sz):(b * bin_sz + bin_sz)] for b in range(n_bins)])

    # Compute fano factor of firing rate for each bin, and the mean fano factor
    ffs = np.var(fr_binned, axis=1) / np.mean(fr_binned, axis=1)
    # ffs[np.isnan(ffs)] = 0 nan's can happen if neuron doesn't spike in a bin
    ff = np.mean(ffs)

    return ff, ffs, fr


def isi_viol(ts, rp=0.002):
    """
    Computes the fraction of isi violations for a unit.

    Parameters
    ----------
    ts : ndarray
        The spike timestamps from which to compute the firing rate.
    rp : float (optional)
        The refractory period (in s).

    Returns
    -------
    frac_isi_viol : float
        The fraction of isi violations.
    n_isi_viol : int
        The number of isi violations.
    isis : ndarray
        The isis.

    See Also
    --------

    Examples
    --------
    1) Get the fraction of isi violations, the total number of isi violations, and the array of
    isis for unit 1.
        >>> unit_idxs = np.where(spks_b['clusters'] == 1)[0]
        >>> ts = spks_b['times'][unit_idxs]
        >>> frac_isi_viol, n_isi_viol, isi = bb.metrics.isi_viol(ts)
    """

    isis = np.diff(ts)
    v = np.where(isis < rp)[0]  # violations
    frac_isi_viol = len(v) / len(ts)
    return frac_isi_viol, len(v), isis


def max_drift(feat):
    """
    Computes the maximum drift (max - min) of a spike feature array.

    Parameters
    ----------
    feat : ndarray
        The spike feature values from which to compute the maximum drift.

    Returns
    -------
    md : float
        The maxmimum drift of the unit.

    See Also
    --------
    cum_drift

    Examples
    --------
    1) Get the maximum depth and amp drift for unit 1.
        >>> unit_idxs = np.where(spks_b['clusters'] == 1)[0]
        >>> depths = spks_b['depths'][unit_idxs]
        >>> amps = spks_b['amps'][unit_idxs]
        >>> depth_md = bb.metrics.max_drift(depths)
        >>> amp_md = bb.metrics.max_drift(amps)
    """

    md = np.max(feat) - np.min(feat)
    return md


def cum_drift(feat):
    """
    Computes the cumulative drift (normalized by the total number of spikes) of a spike feature
    array.

    Parameters
    ----------
    feat : ndarray
        The spike feature values from which to compute the maximum drift.

    Returns
    -------
    cd : float
        The cumulative drift of the unit.

    See Also
    --------
    max_drift

    Examples
    --------
    1) Get the cumulative depth drift for unit 1.
        >>> unit_idxs = np.where(spks_b['clusters'] == 1)[0]
        >>> depths = spks_b['depths'][unit_idxs]
        >>> amps = spks_b['amps'][unit_idxs]
        >>> depth_cd = bb.metrics.cum_drift(depths)
        >>> amp_cd = bb.metrics.cum_drift(amps)
    """

    cd = np.sum(np.abs(np.diff(feat))) / len(feat)
    return cd


def pres_ratio(ts, hist_win=10):
    """
    Computes the presence ratio of spike counts: the number of bins where there is at least one
    spike, over the total number of bins, given a specified bin width.

    Parameters
    ----------
    ts : ndarray
        The spike timestamps from which to compute the presence ratio.
    hist_win : float (optional)
        The time window (in s) to use for computing the presence ratio.

    Returns
    -------
    pr : float
        The presence ratio.
    spks_bins : ndarray
        The number of spks in each bin.

    See Also
    --------
    plot.pres_ratio

    Examples
    --------
    1) Compute the presence ratio for unit 1, given a window of 10 s.
        >>> ts = units_b['times']['1']
        >>> pr, pr_bins = bb.metrics.pres_ratio(ts)
    """

    bins = np.arange(0, ts[-1] + hist_win, hist_win)
    spks_bins, _ = np.histogram(ts, bins)
    pr = len(np.where(spks_bins)[0]) / len(spks_bins)
    return pr, spks_bins


def ptp_over_noise(ephys_file, ts, ch, t=2.0, sr=30000, n_ch_probe=385, dtype='int16', offset=0,
                   car=True):
    """
    For specified channels, for specified timestamps, computes the mean (peak-to-peak amplitudes /
    the MADs of the background noise).

    Parameters
    ----------
    ephys_file : string
        The file path to the binary ephys data.
    ts : ndarray_like
        The timestamps (in s) of the spikes.
    ch : ndarray_like
        The channels on which to extract the waveforms.
    t : numeric (optional)
        The time (in ms) of the waveforms to extract to compute the ptp.
    sr : int (optional)
        The sampling rate (in hz) that the ephys data was acquired at.
    n_ch_probe : int (optional)
        The number of channels of the recording.
    dtype: str (optional)
        The datatype represented by the bytes in `ephys_file`.
    offset: int (optional)
        The offset (in bytes) from the start of `ephys_file`.
    car: bool (optional)
        A flag to perform common-average-referencing before extracting waveforms.

    Returns
    -------
    ptp_sigma : ndarray
        An array containing the mean ptp_over_noise values for the specified `ts` and `ch`.

    Examples
    --------
    1) Compute ptp_over_noise for all spikes on 20 channels around the channel of max amplitude
    for unit 1.
        >>> ts = units_b['times']['1']
        >>> max_ch = max_ch = clstrs_b['channels'][1]
        >>> if max_ch < 10:  # take only channels greater than `max_ch`.
        >>>     ch = np.arange(max_ch, max_ch + 20)
        >>> elif (max_ch + 10) > 385:  # take only channels less than `max_ch`.
        >>>     ch = np.arange(max_ch - 20, max_ch)
        >>> else:  # take `n_c_ch` around `max_ch`.
        >>>     ch = np.arange(max_ch - 10, max_ch + 10)
        >>> p = bb.metrics.ptp_over_noise(ephys_file, ts, ch)
    """

    # Ensure `ch` is ndarray
    ch = np.asarray(ch)
    ch = ch.reshape((ch.size, 1)) if ch.size == 1 else ch

    # Get waveforms.
    wf = bb.io.extract_waveforms(ephys_file, ts, ch, t=t, sr=sr, n_ch_probe=n_ch_probe,
                                 dtype=dtype, offset=offset, car=car)

    # Initialize `mean_ptp` based on `ch`, and compute mean ptp of all spikes for each ch.
    mean_ptp = np.zeros((ch.size,))
    for cur_ch in range(ch.size, ):
        mean_ptp[cur_ch] = np.mean(np.max(wf[:, :, cur_ch], axis=1) -
                                   np.min(wf[:, :, cur_ch], axis=1))

    # Compute MAD for `ch` in chunks.
    s_reader = spikeglx.Reader(ephys_file)
    file_m = s_reader.data  # the memmapped array
    n_chunk_samples = 5e6  # number of samples per chunk
    n_chunks = np.ceil(file_m.shape[0] / n_chunk_samples).astype('int')
    # Get samples that make up each chunk. e.g. `chunk_sample[1] - chunk_sample[0]` are the
    # samples that make up the first chunk.
    chunk_sample = np.arange(0, file_m.shape[0], n_chunk_samples, dtype=int)
    chunk_sample = np.append(chunk_sample, file_m.shape[0])
    # Give time estimate for computing MAD.
    t0 = time.perf_counter()
    stats.median_absolute_deviation(file_m[chunk_sample[0]:chunk_sample[1], ch], axis=0)
    dt = time.perf_counter() - t0
    print('Performing MAD computation. Estimated time is {:.2f} mins.'
          ' ({})'.format(dt * n_chunks / 60, time.ctime()))
    # Compute MAD for each chunk, then take the median MAD of all chunks.
    mad_chunks = np.zeros((n_chunks, ch.size), dtype=np.int16)
    for chunk in range(n_chunks):
        mad_chunks[chunk, :] = stats.median_absolute_deviation(
            file_m[chunk_sample[chunk]:chunk_sample[chunk + 1], ch], axis=0, scale=1)
    print('Done. ({})'.format(time.ctime()))

    # Return `mean_ptp` over `mad`
    mad = np.median(mad_chunks, axis=0)
    ptp_sigma = mean_ptp / mad
    return ptp_sigma


def contamination_est(ts, rp=0.002):
    """
    An estimate of the contamination of the unit (i.e. a pseudo false positive measure) based on
    the number of spikes, number of isi violations, and time between the first and last spike.
    (see Hill et al. (2011) J Neurosci 31: 8699-8705).

    Parameters
    ----------
    ts : ndarray_like
        The timestamps (in s) of the spikes.
    rp : float (optional)
        The refractory period (in s).

    Returns
    -------
    ce : float
        An estimate of the fraction of contamination.

    See Also
    --------
    contamination_est2

    Examples
    --------
    1) Compute contamination estimate for unit 1.
        >>> ts = units_b['times']['1']
        >>> ce = bb.metrics.contamination_est(ts)
    """

    # Get number of spikes, number of isi violations, and time from first to final spike.
    n_spks = len(ts)
    n_isi_viol = np.sum(np.diff(ts) < rp)
    t = ts[-1] - ts[0]

    # `ce` is min of roots of solved quadratic equation.
    c = (t * n_isi_viol) / (2 * rp * n_spks ** 2)  # 3rd term in quadratic
    ce = np.min(np.abs(np.roots([-1, 1, c])))  # solve quadratic
    return ce


def contamination_est2(ts, min_time, max_time, rp=0.002, min_isi=0.0001):
    """
    An estimate of the contamination of the unit (i.e. a pseudo false positive measure) based on
    the number of spikes, number of isi violations, and time between the first and last spike.
    (see Hill et al. (2011) J Neurosci 31: 8699-8705).

    Modified by Dan Denman from cortex-lab/sortingQuality GitHub by Nick Steinmetz.

    Parameters
    ----------
    ts : ndarray_like
        The timestamps (in s) of the spikes.
    min_time : float
        The minimum time (in s) that a potential spike occurred.
    max_time : float
        The maximum time (in s) that a potential spike occurred.
    rp : float (optional)
        The refractory period (in s).
    min_isi : float (optional)
        The minimum interspike-interval (in s) for counting duplicate spikes.

    Returns
    -------
    ce : float
        An estimate of the contamination.
        A perfect unit has a ce = 0
        A unit with some contamination has a ce < 0.5
        A unit with lots of contamination has a ce > 1.0
    num_violations : int
        The total number of isi violations.

    See Also
    --------
    contamination_est

    Examples
    --------
    1) Compute contamination estimate for unit 1, with a minimum isi for counting duplicate
    spikes of 0.1 ms.
        >>> ts = units_b['times']['1']
        >>> ce = bb.metrics.contamination_est2(ts, min_isi=0.0001)
    """

    duplicate_spikes = np.where(np.diff(ts) <= min_isi)[0]

    ts = np.delete(ts, duplicate_spikes + 1)
    isis = np.diff(ts)

    num_spikes = ts.size
    num_violations = np.sum(isis < rp)
    violation_time = 2 * num_spikes * (rp - min_isi)
    total_rate = ts.size / (max_time - min_time)
    violation_rate = num_violations / violation_time
    ce = violation_rate / total_rate

    return ce, num_violations


def quick_unit_metrics(spike_clusters, spike_times, spike_amps, spike_depths,
                       params=METRICS_PARAMS):
    """
    Computes single unit metrics from only the spike times, amplitudes, and depths for a set of
    units.

    Metrics computed:
        num_spikes
        firing_rate
        presence_ratio
        presence_ratio_std
        frac_isi_viol (see `isi_viol`)
        contamination_est (see `contamination_est`)
        contamination_est2 (see `contamination_est2`)
        missed_spikes_est (see `missed_spikes_est`)
        cum_amp_drift (see `cum_drift`)
        max_amp_drift (see `max_drift`)
        cum_depth_drift (see `cum_drift`)
        max_depth_drift (see `max_drift`)

    Parameters
    ----------
    spike_clusters : ndarray_like
        A vector of the unit ids for a set of spikes.
    spike_times : ndarray_like
        A vector of the timestamps for a set of spikes.
    spike_amps : ndarray_like
        A vector of the amplitudes for a set of spikes.
    spike_depths : ndarray_like
        A vector of the depths for a set of spikes.
    params : dict (optional)
        Parameters used for computing some of the metrics in the function:
            'presence_window': float
                The time window (in s) used to look for spikes when computing the presence ratio.
            'refractory_period': float
                The refractory period used when computing isi violations and the contamination
                estimate.
            'min_isi': float
                The minimum interspike-interval (in s) for counting duplicate spikes when computing
                the contamination estimate.
            'spks_per_bin_for_missed_spks_est': int
                The number of spikes per bin used to compute the spike amplitude pdf for a unit,
                when computing the missed spikes estimate.
            'std_smoothing_kernel_for_missed_spks_est': float
                The standard deviation for the gaussian kernel used to compute the spike amplitude
                pdf for a unit, when computing the missed spikes estimate.
            'min_num_bins_for_missed_spks_est': int
                The minimum number of bins used to compute the spike amplitude pdf for a unit,
                when computing the missed spikes estimate.

    Returns
    -------
    r : bunch
        A bunch whose keys are the computed spike metrics.

    Notes
    -----
    This function is called by `ephysqc.unit_metrics_ks2` which is called by `spikes.ks2_to_alf`
    during alf extraction of an ephys dataset in the ibl ephys extraction pipeline.

    Examples
    --------
    1) Compute quick metrics from a ks2 output directory:
        >>> from ibllib.ephys.ephysqc import phy_model_from_ks2_path
        >>> m = phy_model_from_ks2_path(path_to_ks2_out)
        >>> cluster_ids = m.spike_clusters
        >>> ts = m.spike_times
        >>> amps = m.amplitudes
        >>> depths = m.depths
        >>> r = bb.metrics.quick_unit_metrics(cluster_ids, ts, amps, depths)
    """

    cluster_ids = np.arange(np.max(spike_clusters) + 1)
    nclust = cluster_ids.size
    r = Bunch({
        'cluster_id': cluster_ids,
        'num_spikes': np.full((nclust,), np.nan),
        'firing_rate': np.full((nclust,), np.nan),
        'presence_ratio': np.full((nclust,), np.nan),
        'presence_ratio_std': np.full((nclust,), np.nan),
        'frac_isi_viol': np.full((nclust,), np.nan),
        'contamination_est': np.full((nclust,), np.nan),
        'contamination_est2': np.full((nclust,), np.nan),
        'missed_spikes_est': np.full((nclust,), np.nan),
        'cum_amp_drift': np.full((nclust,), np.nan),
        'max_amp_drift': np.full((nclust,), np.nan),
        'cum_depth_drift': np.full((nclust,), np.nan),
        'max_depth_drift': np.full((nclust,), np.nan),
        # could add 'epoch_name' in future:
        # 'epoch_name': np.zeros(nclust, dtype='object'),
    })

    # vectorized computation of basic metrics such as presence ratio and firing rate
    tmin = spike_times[0]
    tmax = spike_times[-1]
    presence_ratio = bincount2D(spike_times, spike_clusters,
                                xbin=params['presence_window'],
                                ybin=cluster_ids, xlim=[tmin, tmax])[0]
    r.presence_ratio = np.sum(presence_ratio > 0, axis=1) / presence_ratio.shape[1]
    r.presence_ratio_std = np.std(presence_ratio, axis=1)
    r.num_spikes = np.sum(presence_ratio, axis=1)
    r.firing_rate = r.num_spikes / (tmax - tmin)

    # loop over each cluster to compute the rest of the metrics
    for ic in np.arange(nclust):
        # slice the spike_times array
        ispikes = spike_clusters == cluster_ids[ic]
        if np.all(~ispikes):  # if this cluster has no spikes, continue
            continue
        ts = spike_times[ispikes]
        amps = spike_amps[ispikes]
        depths = spike_depths[ispikes]

        # compute metrics
        r.frac_isi_viol[ic], _, _ = isi_viol(ts, rp=params['refractory_period'])
        r.contamination_est[ic] = contamination_est(ts, rp=params['refractory_period'])
        r.contamination_est2[ic], _ = contamination_est2(
            ts, tmin, tmax, rp=params['refractory_period'], min_isi=params['min_isi'])
        try:  # this may fail because `missed_spikes_est` requires a min number of spikes
            r.missed_spikes_est[ic], _, _ = missed_spikes_est(
                amps, spks_per_bin=params['spks_per_bin_for_missed_spks_est'],
                sigma=params['std_smoothing_kernel_for_missed_spks_est'],
                min_num_bins=params['min_num_bins_for_missed_spks_est'])
        except AssertionError:
            pass
        r.cum_amp_drift[ic] = cum_drift(amps)
        r.max_amp_drift[ic] = max_drift(amps)
        r.cum_depth_drift[ic] = cum_drift(depths)
        r.max_depth_drift[ic] = max_drift(depths)

    return r



#!/usr/bin/env python
# -*- coding:utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt
import scipy

import ibllib.dsp as dsp


def wiggle(w, fs=1, gain=0.71, color='k', ax=None, fill=True, linewidth=0.5, t0=0, clip=2,
           **kwargs):
    """
    Matplotlib display of wiggle traces

    :param w: 2D array (numpy array dimension nsamples, ntraces)
    :param fs: sampling frequency
    :param gain: display gain
    :param color: ('k') color of traces
    :param ax: (None) matplotlib axes object
    :param fill: (True) fill variable area above 0
    :param t0: (0) timestamp of the first sample
    :return: None
    """
    nech, ntr = w.shape
    tscale = np.arange(nech) / fs
    sf = gain / np.sqrt(dsp.rms(w.flatten()))

    def insert_zeros(trace):
        # Insert zero locations in data trace and tt vector based on linear fit
        # Find zeros
        zc_idx = np.where(np.diff(np.signbit(trace)))[0]
        x1 = tscale[zc_idx]
        x2 = tscale[zc_idx + 1]
        y1 = trace[zc_idx]
        y2 = trace[zc_idx + 1]
        a = (y2 - y1) / (x2 - x1)
        tt_zero = x1 - y1 / a
        # split tt and trace
        tt_split = np.split(tscale, zc_idx + 1)
        trace_split = np.split(trace, zc_idx + 1)
        tt_zi = tt_split[0]
        trace_zi = trace_split[0]
        # insert zeros in tt and trace
        for i in range(len(tt_zero)):
            tt_zi = np.hstack(
                (tt_zi, np.array([tt_zero[i]]), tt_split[i + 1]))
            trace_zi = np.hstack(
                (trace_zi, np.zeros(1), trace_split[i + 1]))
        return trace_zi, tt_zi

    if not ax:
        ax = plt.gca()
    for ntr in range(ntr):
        if fill:
            trace, t_trace = insert_zeros(w[:, ntr] * sf)
            if clip:
                trace = np.maximum(np.minimum(trace, clip), -clip)
            ax.fill_betweenx(t_trace + t0, ntr, trace + ntr,
                             where=trace >= 0,
                             facecolor=color,
                             linewidth=linewidth)
        wplot = np.minimum(np.maximum(w[:, ntr] * sf, -clip), clip)
        ax.plot(wplot + ntr, tscale + t0, color, linewidth=linewidth, **kwargs)

    ax.set_xlim(-1, ntr + 1)
    ax.set_ylim(tscale[0] + t0, tscale[-1] + t0)
    ax.set_ylabel('Time (s)')
    ax.set_xlabel('Trace')
    ax.invert_yaxis()

    return ax


class Density:
    def __init__(self, w, fs=1, cmap='bone', ax=None, **kwargs):
        """
        Matplotlib display of traces as a density display

        :param w: 2D array (numpy array dimension nsamples, ntraces)
        :param fs: sampling frequency (Hz)
        :param ax: axis to plot in
        :return: None
        """
        w = w.reshape(w.shape[0], -1)
        nech, ntr = w.shape
        tscale = np.array([0, nech - 1]) / fs * 1e3
        if ax is None:
            self.figure, ax = plt.subplots()
        else:
            self.figure = ax.get_figure()
        extent = [-0.5, ntr - 0.5, tscale[1], tscale[0]]
        self.im = ax.imshow(w, aspect='auto', cmap=cmap, extent=extent, origin='upper', **kwargs)
        ax.set_ylabel('Time (ms)')
        ax.set_xlabel('Trace')
        self.cid_key = self.figure.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.ax = ax

    def on_key_press(self, event):
        if event.key == 'ctrl+a':
            self.im.set_data(self.im.get_array() * np.sqrt(2))
        elif event.key == 'ctrl+z':
            self.im.set_data(self.im.get_array() / np.sqrt(2))
        else:
            return
        self.figure.canvas.draw()


class Traces:
    def __init__(self, w, fs=1, gain=0.71, color='k', ax=None, linewidth=0.5, t0=0, **kwargs):
        """
        Matplotlib display of traces as a density display

        :param w: 2D array (numpy array dimension nsamples, ntraces)
        :param fs: sampling frequency (Hz)
        :param ax: axis to plot in
        :return: None
        """
        w = w.reshape(w.shape[0], -1)
        nech, ntr = w.shape
        tscale = np.arange(nech) / fs * 1e3
        sf = gain / dsp.rms(w.flatten()) / 2
        if ax is None:
            self.figure, ax = plt.subplots()
        else:
            self.figure = ax.get_figure()
        self.plot = ax.plot(w * sf + np.arange(ntr), tscale + t0, color,
                            linewidth=linewidth, **kwargs)
        ax.set_xlim(-1, ntr + 1)
        ax.set_ylim(tscale[0] + t0, tscale[-1] + t0)
        ax.set_ylabel('Time (ms)')
        ax.set_xlabel('Trace')
        ax.invert_yaxis()
        self.cid_key = self.figure.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.ax = ax

    def on_key_press(self, event):
        if event.key == 'ctrl+a':
            for i, l in enumerate(self.plot):
                l.set_xdata((l.get_xdata() - i) * np.sqrt(2) + i)
        elif event.key == 'ctrl+z':
            for i, l in enumerate(self.plot):
                l.set_xdata((l.get_xdata() - i) / np.sqrt(2) + i)
        else:
            return
        self.figure.canvas.draw()


def squares(tscale, polarity, ax=None, yrange=[-1, 1], **kwargs):
    """
    Matplotlib display of rising and falling fronts in a square-wave pattern

    :param tscale: time of indices of fronts
    :param polarity: polarity of front (1: rising, -1:falling)
    :param ax: matplotlib axes object
    :return: None
    """
    if not ax:
        ax = plt.gca()
    isort = np.argsort(tscale)
    tscale = tscale[isort]
    polarity = polarity[isort]
    f = np.tile(polarity, (2, 1))
    t = np.concatenate((tscale, np.r_[tscale[1:], tscale[-1]])).reshape(2, f.shape[1])
    ydata = f.transpose().ravel()
    ydata = (ydata + 1) / 2 * (yrange[1] - yrange[0]) + yrange[0]
    ax.plot(t.transpose().ravel(), ydata, **kwargs)


def vertical_lines(x, ymin=0, ymax=1, ax=None, **kwargs):
    """
    From a x vector, draw separate vertical lines at each x location ranging from ymin to ymax

    :param x: numpy array vector of x values where to display lnes
    :param ymin: lower end of the lines (scalar)
    :param ymax: higher end of the lines (scalar)
    :param ax: (optional) matplotlib axis instance
    :return: None
    """
    x = np.tile(x, (3, 1))
    x[2, :] = np.nan
    y = np.zeros_like(x)
    y[0, :] = ymin
    y[1, :] = ymax
    y[2, :] = np.nan
    if not ax:
        ax = plt.gca()
    ax.plot(x.T.flatten(), y.T.flatten(), **kwargs)


def spectrum(w, fs, smooth=None, unwrap=True, axis=0, **kwargs):
    """
    Display spectral density of a signal along a given dimension
    spectrum(w, fs)
    :param w: signal
    :param fs: sampling frequency (Hz)
    :param smooth: (None) frequency samples to smooth over
    :param unwrap: (True) unwraps the phase specrum
    :param axis: axis on which to compute the FFT
    :param kwargs: plot arguments to be passed to matplotlib
    :return: matplotlib axes
    """
    axis = 0
    smooth = None
    unwrap = True

    ns = w.shape[axis]
    fscale = dsp.fscale(ns, 1 / fs, one_sided=True)
    W = scipy.fft.rfft(w, axis=axis)
    amp = 20 * np.log10(np.abs(W))
    phi = np.angle(W)

    if unwrap:
        phi = np.unwrap(phi)

    if smooth:
        nf = np.round(smooth / fscale[1] / 2) * 2 + 1
        amp = dsp.smooth.mwa(amp, nf)
        phi = dsp.smooth.mwa(phi, nf)

    fig, ax = plt.subplots(2, 1, sharex=True)
    ax[0].plot(fscale, amp, **kwargs)
    ax[1].plot(fscale, phi, **kwargs)

    ax[0].set_title('Spectral Density (dB rel to amplitude.Hz^-0.5)')
    ax[0].set_ylabel('Amp (dB)')
    ax[1].set_ylabel('Phase (rad)')
    ax[1].set_xlabel('Frequency (Hz)')
    return ax


def color_cycle(ind=None):
    """
    Gets the matplotlib color-cycle as RGB numpy array of floats between 0 and 1
    :return:
    """
    # import matplotlib as mpl
    # c = np.uint32(np.array([int(c['color'][1:], 16) for c in mpl.rcParams['axes.prop_cycle']]))
    # c = np.double(np.flip(np.reshape(c.view(np.uint8), (c.size, 4))[:, :3], 1)) / 255
    c = np.array([[0.12156863, 0.46666667, 0.70588235],
                  [1., 0.49803922, 0.05490196],
                  [0.17254902, 0.62745098, 0.17254902],
                  [0.83921569, 0.15294118, 0.15686275],
                  [0.58039216, 0.40392157, 0.74117647],
                  [0.54901961, 0.3372549, 0.29411765],
                  [0.89019608, 0.46666667, 0.76078431],
                  [0.49803922, 0.49803922, 0.49803922],
                  [0.7372549, 0.74117647, 0.13333333],
                  [0.09019608, 0.74509804, 0.81176471]])
    if ind is None:
        return c
    else:
        return tuple(c[ind % c.shape[0], :])


if __name__ == "__main__":
    w = np.random.rand(500, 40) - 0.5
    wiggle(w, fs=30000)
    Traces(w, fs=30000, color='r')



import logging

import numpy as np

from ibllib.io.extractors.training_trials import (
    StimOnOffFreezeTimes, Choice, FeedbackType, Intervals, StimOnTriggerTimes, StimOnTimes,
    StimOffTriggerTimes, StimFreezeTriggerTimes, GoCueTriggerTimes, GoCueTimes,
    ErrorCueTriggerTimes, RewardVolume, ResponseTimes, FeedbackTimes, ItiInTimes,
    ProbabilityLeft, run_extractor_classes  # ContrastLR
)
from ibllib.io.extractors.training_wheel import Wheel
from oneibl.one import ONE
from ibllib.io.extractors.ephys_fpga import (
    _get_pregenerated_events, _get_main_probe_sync, bpod_fpga_sync, FpgaTrials
)
import ibllib.io.raw_data_loaders as raw


class TaskQCExtractor(object):
    def __init__(self, session_path, lazy=False, one=None, download_data=False, bpod_only=False):
        """
        A class for extracting the task data required to perform task quality control
        :param session_path: a valid session path
        :param lazy: if True, the data are not extracted immediately
        :param one: an instance of ONE, used to download the raw data if download_data is True
        :param download_data: if True, any missing raw data is downloaded via ONE
        :param bpod_only: extract from from raw Bpod data only, even for FPGA sessions
        """
        self.session_path = session_path
        self.one = one or ONE()
        self.log = logging.getLogger("ibllib")

        self.data = None
        self.settings = None
        self.raw_data = None
        self.frame_ttls = self.audio_ttls = None
        self.type = None
        self.wheel_encoding = None
        self.bpod_only = bpod_only

        if download_data:
            self._ensure_required_data()

        if not lazy:
            self.load_raw_data()
            self.extract_data()

    def _ensure_required_data(self):
        """
        Attempt to download any required raw data if missing, and raise exception if any data are
        missing.
        :return:
        """
        dstypes = [
            "_iblrig_taskData.raw",
            "_iblrig_taskSettings.raw",
            "_iblrig_encoderPositions.raw",
            "_iblrig_encoderEvents.raw",
            "_iblrig_stimPositionScreen.raw",
            "_iblrig_syncSquareUpdate.raw",
            "_iblrig_encoderTrialInfo.raw",
            "_iblrig_ambientSensorData.raw",
        ]
        eid = self.one.eid_from_path(self.session_path)
        # Ensure we have the settings
        settings = self.one.load(eid, ["_iblrig_taskSettings.raw"], download_only=True)
        if settings and raw.get_session_extractor_type(self.session_path) == 'ephys':
            dstypes.extend(['_spikeglx_sync.channels',
                            '_spikeglx_sync.polarities',
                            '_spikeglx_sync.times',
                            'ephysData.raw.meta',
                            'ephysData.raw.wiring'])
        self.log.info(f"Downloading data for session {eid}")
        files = self.one.load(eid, dataset_types=dstypes, download_only=True)
        missing = [True for _ in dstypes] if not files else [x is None for x in files]
        if self.session_path is None or all(missing):
            self.lazy = True
            self.log.error("Data not found on server, can't calculate QC.")
        elif any(missing):
            self.log.warning(
                f"Missing some datasets for session {eid} in path {self.session_path}"
            )

    def load_raw_data(self):
        """
        Loads the TTLs, raw task data and task settings
        :return:
        """
        self.log.info(f"Loading raw data from {self.session_path}")
        self.type = self.type or raw.get_session_extractor_type(self.session_path)
        self.settings, self.raw_data = raw.load_bpod(self.session_path)
        # Fetch the TTLs for the photodiode and audio
        if self.type != 'ephys' or self.bpod_only is True:  # Extract from Bpod
            ttls = raw.load_bpod_fronts(self.session_path, data=self.raw_data)
        else:  # Extract from FPGA
            sync, chmap = _get_main_probe_sync(self.session_path)

            def channel_events(name):
                """Fetches the polarities and times for a given channel"""
                keys = ('polarities', 'times')
                mask = sync['channels'] == chmap[name]
                return dict(zip(keys, (sync[k][mask] for k in keys)))

            ttls = [channel_events(ch) for ch in ('frame2ttl', 'audio')]
        self.frame_ttls, self.audio_ttls = ttls

    def extract_data(self, partial=False):
        """Extracts and loads behaviour data for QC
        NB: partial extraction when bpod_only sttricbute is False requires intervals and
        intervals_bpod to be assigned to the data attribute before calling this function.
        :param partial: If True, extracts only the required data that aren't usually saved to ALFs
        :return:
        """
        self.log.info(f"Extracting session: {self.session_path}")
        self.type = self.type or raw.get_session_extractor_type(self.session_path)
        self.wheel_encoding = 'X4' if (self.type == 'ephys' and not self.bpod_only) else 'X1'

        # Partial extraction for FPGA sessions only worth it if intervals already extracted and
        # assigned to the data attribute
        data_assigned = self.data and {'intervals', 'intervals_bpod'}.issubset(self.data)
        if partial and self.type == 'ephys' and not self.bpod_only and not data_assigned:
            partial = False  # Requires intervals for converting to FPGA time

        if not self.raw_data:
            self.load_raw_data()

        # Signals and parameters not usually saved to file, available for all task types
        extractors = [
            StimOnTriggerTimes, StimOffTriggerTimes, StimOnOffFreezeTimes,
            StimFreezeTriggerTimes, ErrorCueTriggerTimes, ItiInTimes]

        # Extract the data that are usually saved to file
        if not partial:
            if self.type == 'ephys' and not self.bpod_only:
                extractors.append(FpgaTrials)
            else:
                extractors.extend([
                    Choice, FeedbackType, Intervals, StimOnTimes, GoCueTriggerTimes, Wheel,
                    GoCueTimes, RewardVolume, ResponseTimes, FeedbackTimes, ProbabilityLeft])
                # if type == 'biased':
                #     # FIXME ContrastLR fails on old sessions (contrast is a float, not a dict)
                #     extractors.append(ContrastLR)

        # Run behaviour extractors
        kwargs = dict(save=False, bpod_trials=self.raw_data, settings=self.settings)
        data, _ = run_extractor_classes(extractors, session_path=self.session_path, **kwargs)

        # Extract some parameters
        if self.type == 'ephys':
            # For ephys sessions extract quiescence and phase from pre-generated file
            data.update(_get_pregenerated_events(self.raw_data, self.settings))

            if not self.bpod_only:
                # Get the extracted intervals for sync.  For partial ephys extraction attempt to
                # get intervals from data attribute.
                intervals, intervals_bpod = [data[key] if key in data else self.data[key]
                                             for key in ('intervals', 'intervals_bpod')]
                # We need to sync the extra extracted data to FPGA time
                # 0.5s iti already removed during extraction so we set duration to 0 here
                ibpod, _, bpod2fpga = bpod_fpga_sync(intervals_bpod, intervals, iti_duration=0)
                # These fields have to be re-synced
                sync_fields = ['stimOnTrigger_times', 'stimOffTrigger_times', 'stimFreeze_times',
                               'stimFreezeTrigger_times', 'errorCueTrigger_times', 'itiIn_times']
                # build trials output
                data.update({k: bpod2fpga(data[k][ibpod]) for k in sync_fields})
        else:
            data['quiescence'] = np.array([t['quiescent_period'] for t in self.raw_data])
            data['position'] = np.array([t['position'] for t in self.raw_data])
            self.wheel_encoding = 'X1'

        # Update the data attribute with extracted data
        if self.data:
            self.data.update(data)
            self.rename_data(self.data)
        else:
            self.data = data if partial else self.rename_data(data)

    @staticmethod
    def rename_data(data):
        """Rename the extracted data dict for use with TaskQC
        Splits 'feedback_times' to 'errorCue_times' and 'valveOpen_times'.  Also adds 'outcome' and
        'correct'.
        NB: The data is not copied before making changes
        :param data: A dict of task data returned by the task extractors
        :return: the same dict after modifying the keys
        """
        # get valve_time and errorCue_times from feedback_times
        correct = data['feedbackType'] > 0
        errorCue_times = data["feedback_times"].copy()
        valveOpen_times = data["feedback_times"].copy()
        errorCue_times[correct] = np.nan
        valveOpen_times[~correct] = np.nan
        data.update(
            {"errorCue_times": errorCue_times,
             "valveOpen_times": valveOpen_times,
             "correct": correct}
        )
        data["outcome"] = data["feedbackType"].copy()
        data["outcome"][data["choice"] == 0] = 0
        return data



'''
Processes data from one form into another, e.g. taking spike times and binning them into
non-overlapping bins and convolving spike times with a gaussian kernel.
'''

import numpy as np
import pandas as pd
from scipy import interpolate
from brainbox import core


def sync(dt, times=None, values=None, timeseries=None, offsets=None, interp='zero',
         fillval=np.nan):
    """
    Function for resampling a single or multiple time series to a single, evenly-spaced, delta t
    between observations. Uses interpolation to find values.

    Can be used on raw numpy arrays of timestamps and values using the 'times' and 'values' kwargs
    and/or on brainbox.core.TimeSeries objects passed to the 'timeseries' kwarg. If passing both
    TimeSeries objects and numpy arrays, the offsets passed should be for the TS objects first and
    then the numpy arrays.

    Uses scipy's interpolation library to perform interpolation.
    See scipy.interp1d for more information regarding interp and fillval parameters.

    :param dt: Separation of points which the output timeseries will be sampled at
    :type dt: float
    :param timeseries: A group of time series to perform alignment or a single time series.
        Must have time stamps.
    :type timeseries: tuple of TimeSeries objects, or a single TimeSeries object.
    :param times: time stamps for the observations in 'values']
    :type times: np.ndarray or list of np.ndarrays
    :param values: observations corresponding to the timestamps in 'times'
    :type values: np.ndarray or list of np.ndarrays
    :param offsets: tuple of offsets for time stamps of each time series. Offsets for passed
        TimeSeries objects first, then offsets for passed numpy arrays. defaults to None
    :type offsets: tuple of floats, optional
    :param interp: Type of interpolation to use. Refer to scipy.interpolate.interp1d for possible
        values, defaults to np.nan
    :type interp: str
    :param fillval: Fill values to use when interpolating outside of range of data. See interp1d
        for possible values, defaults to np.nan
    :return: TimeSeries object with each row representing synchronized values of all
        input TimeSeries. Will carry column names from input time series if all of them have column
        names.
    """
    #########################################
    # Checks on inputs and input processing #
    #########################################

    # Initialize a list to contain times/values pairs if no TS objs are passed
    if timeseries is None:
        timeseries = []
    # If a single time series is passed for resampling, wrap it in an iterable
    elif isinstance(timeseries, core.TimeSeries):
        timeseries = [timeseries]
    # Yell at the user if they try to pass stuff to timeseries that isn't a TimeSeries object
    elif not all([isinstance(ts, core.TimeSeries) for ts in timeseries]):
        raise TypeError('All elements of \'timeseries\' argument must be brainbox.core.TimeSeries '
                        'objects. Please uses \'times\' and \'values\' for np.ndarray args.')
    # Check that if something is passed to times or values, there is a corresponding equal-length
    # argument for the other element.
    if (times is not None) or (values is not None):
        if len(times) != len(values):
            raise ValueError('\'times\' and \'values\' must have the same number of elements.')
        if type(times[0]) is np.ndarray:
            if not all([t.shape == v.shape for t, v in zip(times, values)]):
                raise ValueError('All arrays in \'times\' must match the shape of the'
                                 ' corresponding entry in \'values\'.')
            # If all checks are passed, convert all times and values args into TimeSeries objects
            timeseries.extend([core.TimeSeries(t, v) for t, v in zip(times, values)])
        else:
            # If times and values are only numpy arrays and lists of arrays, pair them and add
            timeseries.append(core.TimeSeries(times, values))

    # Adjust each timeseries by the associated offset if necessary then load into a list
    if offsets is not None:
        tstamps = [ts.times + os for ts, os in zip(timeseries, offsets)]
    else:
        tstamps = [ts.times for ts in timeseries]
    # If all input timeseries have column names, put them together for the output TS
    if all([ts.columns is not None for ts in timeseries]):
        colnames = []
        for ts in timeseries:
            colnames.extend(ts.columns)
    else:
        colnames = None

    #################
    # Main function #
    #################

    # Get the min and max values for all timeseries combined after offsetting
    tbounds = np.array([(np.amin(ts), np.amax(ts)) for ts in tstamps])
    if not np.all(np.isfinite(tbounds)):
        # If there is a np.inf or np.nan in the time stamps for any of the timeseries this will
        # break any further code so we check for all finite values and throw an informative error.
        raise ValueError('NaN or inf encountered in passed timeseries.\
                          Please either drop or fill these values.')
    tmin, tmax = np.amin(tbounds[:, 0]), np.amax(tbounds[:, 1])
    if fillval == 'extrapolate':
        # If extrapolation is enabled we can ensure we have a full coverage of the data by
        # extending the t max to be an whole integer multiple of dt above tmin.
        # The 0.01% fudge factor is to account for floating point arithmetic errors.
        newt = np.arange(tmin, tmax + 1.0001 * (dt - (tmax - tmin) % dt), dt)
    else:
        newt = np.arange(tmin, tmax, dt)
    tsinterps = [interpolate.interp1d(ts.times, ts.values, kind=interp, fill_value=fillval, axis=0)
                 for ts in timeseries]
    syncd = core.TimeSeries(newt, np.hstack([tsi(newt) for tsi in tsinterps]), columns=colnames)
    return syncd


def bincount2D(x, y, xbin=0, ybin=0, xlim=None, ylim=None, weights=None):
    """
    Computes a 2D histogram by aggregating values in a 2D array.

    :param x: values to bin along the 2nd dimension (c-contiguous)
    :param y: values to bin along the 1st dimension
    :param xbin:
        scalar: bin size along 2nd dimension
        0: aggregate according to unique values
        array: aggregate according to exact values (count reduce operation)
    :param ybin:
        scalar: bin size along 1st dimension
        0: aggregate according to unique values
        array: aggregate according to exact values (count reduce operation)
    :param xlim: (optional) 2 values (array or list) that restrict range along 2nd dimension
    :param ylim: (optional) 2 values (array or list) that restrict range along 1st dimension
    :param weights: (optional) defaults to None, weights to apply to each value for aggregation
    :return: 3 numpy arrays MAP [ny,nx] image, xscale [nx], yscale [ny]
    """
    # if no bounds provided, use min/max of vectors
    if xlim is None:
        xlim = [np.min(x), np.max(x)]
    if ylim is None:
        ylim = [np.min(y), np.max(y)]

    def _get_scale_and_indices(v, bin, lim):
        # if bin is a nonzero scalar, this is a bin size: create scale and indices
        if np.isscalar(bin) and bin != 0:
            scale = np.arange(lim[0], lim[1] + bin / 2, bin)
            ind = (np.floor((v - lim[0]) / bin)).astype(np.int64)
        # if bin == 0, aggregate over unique values
        else:
            scale, ind = np.unique(v, return_inverse=True)
        return scale, ind

    xscale, xind = _get_scale_and_indices(x, xbin, xlim)
    yscale, yind = _get_scale_and_indices(y, ybin, ylim)
    # aggregate by using bincount on absolute indices for a 2d array
    nx, ny = [xscale.size, yscale.size]
    ind2d = np.ravel_multi_index(np.c_[yind, xind].transpose(), dims=(ny, nx))
    r = np.bincount(ind2d, minlength=nx * ny, weights=weights).reshape(ny, nx)

    # if a set of specific values is requested output an array matching the scale dimensions
    if not np.isscalar(xbin) and xbin.size > 1:
        _, iout, ir = np.intersect1d(xbin, xscale, return_indices=True)
        _r = r.copy()
        r = np.zeros((ny, xbin.size))
        r[:, iout] = _r[:, ir]
        xscale = xbin

    if not np.isscalar(ybin) and ybin.size > 1:
        _, iout, ir = np.intersect1d(ybin, yscale, return_indices=True)
        _r = r.copy()
        r = np.zeros((ybin.size, r.shape[1]))
        r[iout, :] = _r[ir, :]
        yscale = ybin

    return r, xscale, yscale


def bin_spikes(spikes, binsize, interval_indices=False):
    """
    Wrapper for bincount2D which is intended to take in a TimeSeries object of spike times
    and cluster identities and spit out spike counts in bins of a specified width binsize, also in
    another TimeSeries object. Can either return a TS object with each row labeled with the
    corresponding interval or the value of the left edge of the bin.

    :param spikes: Spike times and cluster identities of sorted spikes
    :type spikes: TimeSeries object with \'clusters\' column and timestamps
    :param binsize: Width of the non-overlapping bins in which to bin spikes
    :type binsize: float
    :param interval_indices: Whether to use intervals as the time stamps for binned spikes, rather
        than the left edge value of the bins, defaults to False
    :type interval_indices: bool, optional
    :return: Object with 2D array of shape T x N, for T timesteps and N clusters, and the
        associated time stamps.
    :rtype: TimeSeries object
    """
    if type(spikes) is not core.TimeSeries:
        raise TypeError('Input spikes need to be in TimeSeries object format')

    if not hasattr(spikes, 'clusters'):
        raise AttributeError('Input spikes need to have a clusters attribute. Make sure you set '
                             'columns=(\'clusters\',)) when constructing spikes.')

    rates, tbins, clusters = bincount2D(spikes.times, spikes.clusters, binsize)
    if interval_indices:
        intervals = pd.interval_range(tbins[0], tbins[-1], freq=binsize, closed='left')
        return core.TimeSeries(times=intervals, values=rates.T[:-1], columns=clusters)
    else:
        return core.TimeSeries(times=tbins, values=rates.T, columns=clusters)


def get_units_bunch(spks_b, *args):
    '''
    Returns a bunch, where the bunch keys are keys from `spks` with labels of spike information
    (e.g. unit IDs, times, features, etc.), and the values for each key are arrays with values for
    each unit: these arrays are ordered and can be indexed by unit id.

    Parameters
    ----------
    spks_b : bunch
        A spikes bunch containing fields with spike information (e.g. unit IDs, times, features,
        etc.) for all spikes.
    features : list of strings (optional positional arg)
        A list of names of labels of spike information (which must be keys in `spks`) that specify
        which labels to return as keys in `units`. If not provided, all keys in `spks` are returned
        as keys in `units`.

    Returns
    -------
    units_b : bunch
        A bunch with keys of labels of spike information (e.g. cluster IDs, times, features, etc.)
        whose values are arrays that hold values for each unit. The arrays for each key are ordered
        by unit ID.

    Examples
    --------
    1) Create a units bunch given a spikes bunch, and get the amps for unit #4 from the units
    bunch.
        >>> import brainbox as bb
        >>> import alf.io as aio
        >>> import ibllib.ephys.spikes as e_spks
        (*Note, if there is no 'alf' directory, make 'alf' directory from 'ks2' output directory):
        >>> e_spks.ks2_to_alf(path_to_ks_out, path_to_alf_out)
        >>> spks_b = aio.load_object(path_to_alf_out, 'spikes')
        >>> units_b = bb.processing.get_units_bunch(spks_b)
        # Get amplitudes for unit 4.
        >>> amps = units_b['amps']['4']

    TODO add computation time estimate?
    '''

    # Initialize `units`
    units_b = core.Bunch()
    # Get the keys to return for `units`:
    if not args:
        feat_keys = list(spks_b.keys())
    else:
        feat_keys = args[0]
    # Get unit id for each spike and number of units. *Note: `n_units` might not equal `len(units)`
    # because some clusters may be empty (due to a "wontfix" bug in ks2).
    spks_unit_id = spks_b['clusters']
    n_units = np.max(spks_unit_id)
    units = np.unique(spks_b['clusters'])
    # For each key in `units`, iteratively get each unit's values and add as a key to a bunch,
    # `feat_bunch`. After iterating through all units, add `feat_bunch` as a key to `units`:
    for feat in feat_keys:
        # Initialize `feat_bunch` with a key for each unit.
        feat_bunch = core.Bunch((str(unit), np.array([])) for unit in np.arange(n_units))
        for unit in units:
            unit_idxs = np.where(spks_unit_id == unit)[0]
            feat_bunch[str(unit)] = spks_b[feat][unit_idxs]
        units_b[feat] = feat_bunch
    return units_b


def filter_units(units_b, t, **kwargs):
    '''
    Filters units according to some parameters. **kwargs are the keyword parameters used to filter
    the units.

    Parameters
    ----------
    units_b : bunch
        A bunch with keys of labels of spike information (e.g. cluster IDs, times, features, etc.)
        whose values are arrays that hold values for each unit. The arrays for each key are ordered
        by unit ID.
    t : float
        Duration of time over which to calculate the firing rate and false positive rate.

    Keyword Parameters
    ------------------
    min_amp : float
        The minimum mean amplitude (in V) of the spikes in the unit. Default value is 50e-6.
    min_fr : float
        The minimum firing rate (in Hz) of the unit. Default value is 0.5.
    max_fpr : float
        The maximum false positive rate of the unit (using the fp formula in Hill et al. (2011)
        J Neurosci 31: 8699-8705). Default value is 0.2.
    rp : float
        The refractory period (in s) of the unit. Used to calculate `max_fp`. Default value is
        0.002.

    Returns
    -------
    filt_units : ndarray
        The ids of the filtered units.

    See Also
    --------
    get_units_bunch

    Examples
    --------
    1) Filter units according to the default parameters.
        >>> import brainbox as bb
        >>> import alf.io as aio
        >>> import ibllib.ephys.spikes as e_spks
        (*Note, if there is no 'alf' directory, make 'alf' directory from 'ks2' output directory):
        >>> e_spks.ks2_to_alf(path_to_ks_out, path_to_alf_out)
        # Get a spikes bunch, units bunch, and filter the units.
        >>> spks_b = aio.load_object(path_to_alf_out, 'spikes')
        >>> units_b = bb.processing.get_units_bunch(spks_b, ['times', 'amps', 'clusters'])
        >>> T = spks_b['times'][-1] - spks_b['times'][0]
        >>> filtered_units = bb.processing.filter_units(units_b, T)

    2) Filter units with no minimum amplitude, a minimum firing rate of 1 Hz, and a max false
    positive rate of 0.2, given a refractory period of 2 ms.
        >>> filtered_units  = bb.processing.filter_units(units_b, T, min_amp=0, min_fr=1)

    TODO: `units_b` input arg could eventually be replaced by `clstrs_b` if the required metrics
          are in `clstrs_b['metrics']`
    '''

    # Set params
    params = {'min_amp': 50e-6, 'min_fr': 0.5, 'max_fpr': 0.2, 'rp': 0.002}  # defaults
    params.update(kwargs)  # update from **kwargs

    # Iteratively filter the units for each filter param #
    # -------------------------------------------------- #
    units = np.asarray(list(units_b.amps.keys()))
    # Remove empty clusters
    empty_cl = np.where([len(units_b.amps[unit]) == 0 for unit in units])[0]
    filt_units = np.delete(units, empty_cl)
    for param in params.keys():
        if param == 'min_amp':  # return units above with amp > `'min_amp'`
            mean_amps = np.asarray([np.mean(units_b.amps[unit]) for unit in filt_units])
            filt_idxs = np.where(mean_amps > params['min_amp'])[0]
            filt_units = filt_units[filt_idxs]
        elif param == 'min_fr':  # return units with fr > `'min_fr'`
            fr = np.asarray([len(units_b.amps[unit]) /
                            (units_b.times[unit][-1] - units_b.times[unit][0])
                            for unit in filt_units])
            filt_idxs = np.where(fr > params['min_fr'])[0]
            filt_units = filt_units[filt_idxs]
        elif param == 'max_fpr':  # return units with fpr < `'max_fpr'`
            fpr = np.zeros_like(filt_units, dtype='float')
            for i, unit in enumerate(filt_units):
                n_spks = len(units_b.amps[unit])
                n_isi_viol = len(np.where(np.diff(units_b.times[unit]) < params['rp'])[0])
                # fpr is min of roots of solved quadratic equation (Hill, et al. 2011).
                c = (t * n_isi_viol) / (2 * params['rp'] * n_spks**2)  # 3rd term in quadratic
                fpr[i] = np.min(np.abs(np.roots([-1, 1, c])))  # solve quadratic
            filt_idxs = np.where(fpr < params['max_fpr'])[0]
            filt_units = filt_units[filt_idxs]
    return filt_units.astype(int)



## Init
from oneibl.one import ONE
from ibllib.misc import pprint
one = ONE(base_url='https://test.alyx.internationalbrainlab.org', username='test_user',
          password='TapetesBloc18')

## Find an experiment
eid = one.search(users='olivier', date_range=['2018-08-24', '2018-08-24'])
pprint(eid)
one.search_terms()

## List dataset types for a session
eid = 'cf264653-2deb-44cb-aa84-89b82507028a'
one.list(eid)
## More Info about a session
d = one.list(eid, 'All')

## Get More Info about datasets
d = one.list(eid, details=True)
print(d)

print(d)
## List #1
one.list(None, 'dataset-types')
one.list(None, 'users')
one.list(None, 'subjects')

## Load #1
dataset_types = ['clusters.templateWaveforms', 'clusters.probes', 'clusters.depths']
eid = 'cf264653-2deb-44cb-aa84-89b82507028a'
wf, pr, d = one.load(eid, dataset_types=dataset_types)

## Load #2
my_data = one.load(eid, dataset_types=dataset_types, dclass_output=True)
from ibllib.misc import pprint
pprint(str(my_data.local_path))
pprint(my_data.dataset_type)

## Load everything
eid = one.search(subjects='flowers')
my_data = one.load(eid[0])
pprint(my_data.dataset_type)

## Load
eid = 'cf264653-2deb-44cb-aa84-89b82507028a'
dataset_types = ['clusters.probes', 'thisDataset.IveJustMadeUp', 'clusters.depths']
t, empty, cl = one.load(eid, dataset_types=dataset_types)



## Search users
eid = one.search(users=['olivier'])

eid = one.search(users=['nbonacchi', 'olivier'])
# with details
eid, session_details = one.search(users=['test_user', 'olivier'], details=True)
pprint(session_details)

## Search by date
eid = one.search(users='olivier', date_range=['2018-08-24', '2018-08-24'])





import time
import numpy as np
# (Previously required `os.path` to get file info before memmapping)
# import os.path as op
from ibllib.io import spikeglx


def extract_waveforms(ephys_file, ts, ch, t=2.0, sr=30000, n_ch_probe=385, dtype='int16',
                      offset=0, car=True):
    '''
    Extracts spike waveforms from binary ephys data file, after (optionally)
    common-average-referencing (CAR) spatial noise.

    Parameters
    ----------
    ephys_file : string
        The file path to the binary ephys data.
    ts : ndarray_like
        The timestamps (in s) of the spikes.
    ch : ndarray_like
        The channels on which to extract the waveforms.
    t : numeric (optional)
        The time (in ms) of each returned waveform.
    sr : int (optional)
        The sampling rate (in hz) that the ephys data was acquired at.
    n_ch_probe : int (optional)
        The number of channels of the recording.
    dtype: str (optional)
        The datatype represented by the bytes in `ephys_file`.
    offset: int (optional)
        The offset (in bytes) from the start of `ephys_file`.
    car: bool (optional)
        A flag to perform CAR before extracting waveforms.

    Returns
    -------
    waveforms : ndarray
        An array of shape (#spikes, #samples, #channels) containing the waveforms.

    Examples
    --------
    1) Extract all the waveforms for unit1 with and without CAR.
        >>> import numpy as np
        >>> import brainbox as bb
        >>> import alf.io as aio
        >>> import ibllib.ephys.spikes as e_spks
        (*Note, if there is no 'alf' directory, make 'alf' directory from 'ks2' output directory):
        >>> e_spks.ks2_to_alf(path_to_ks_out, path_to_alf_out)
        # Get a clusters bunch and a units bunch from a spikes bunch from an alf directory.
        >>> clstrs_b = aio.load_object(path_to_alf_out, 'clusters')
        >>> spks_b = aio.load_object(path_to_alf_out, 'spikes')
        >>> units_b = bb.processing.get_units_bunch(spks, ['times'])
        # Get the timestamps and 20 channels around the max amp channel for unit1, and extract the
        # two sets of waveforms.
        >>> ts = units_b['times']['1']
        >>> max_ch = max_ch = clstrs_b['channels'][1]
        >>> if max_ch < 10:  # take only channels greater than `max_ch`.
        >>>     ch = np.arange(max_ch, max_ch + 20)
        >>> elif (max_ch + 10) > 385:  # take only channels less than `max_ch`.
        >>>     ch = np.arange(max_ch - 20, max_ch)
        >>> else:  # take `n_c_ch` around `max_ch`.
        >>>     ch = np.arange(max_ch - 10, max_ch + 10)
        >>> wf = bb.io.extract_waveforms(path_to_ephys_file, ts, ch, car=False)
        >>> wf_car = bb.io.extract_waveforms(path_to_ephys_file, ts, ch, car=True)
    '''

    # (Previously memmaped the file manually, but now use `spikeglx.Reader`)
    # item_bytes = np.dtype(dtype).itemsize
    # n_samples = (op.getsize(ephys_file) - offset) // (item_bytes * n_ch_probe)
    # file_m = np.memmap(ephys_file, shape=(n_samples, n_ch_probe), dtype=dtype, mode='r')

    # Get memmapped array of `ephys_file`
    s_reader = spikeglx.Reader(ephys_file)
    file_m = s_reader.data  # the memmapped array
    n_wf_samples = np.int(sr / 1000 * (t / 2))  # number of samples to return on each side of a ts
    ts_samples = np.array(ts * sr).astype(int)  # the samples corresponding to `ts`
    t_sample_first = ts_samples[0] - n_wf_samples

    # Exception handling for impossible channels
    ch = np.asarray(ch)
    ch = ch.reshape((ch.size, 1)) if ch.size == 1 else ch
    if np.any(ch < 0) or np.any(ch > n_ch_probe):
        raise Exception('At least one specified channel number is impossible. The minimum channel'
                        ' number was {}, and the maximum channel number was {}. Check specified'
                        ' channel numbers and try again.'.format(np.min(ch), np.max(ch)))

    if car:  # compute spatial noise in chunks
        # see https://github.com/int-brain-lab/iblenv/issues/5
        raise NotImplementedError("CAR option is not available")

    # Initialize `waveforms`, extract waveforms from `file_m`, and CAR.
    waveforms = np.zeros((len(ts), 2 * n_wf_samples, ch.size))
    # Give time estimate for extracting waveforms.
    t0 = time.perf_counter()
    for i in range(5):
        waveforms[i, :, :] = \
            file_m[i * n_wf_samples * 2 + t_sample_first:
                   i * n_wf_samples * 2 + t_sample_first + n_wf_samples * 2, ch].reshape(
                       (n_wf_samples * 2, ch.size))
    dt = time.perf_counter() - t0
    print('Performing waveform extraction. Estimated time is {:.2f} mins. ({})'
          .format(dt * len(ts) / 60 / 5, time.ctime()))
    for spk, _ in enumerate(ts):  # extract waveforms
        spk_ts_sample = ts_samples[spk]
        spk_samples = np.arange(spk_ts_sample - n_wf_samples, spk_ts_sample + n_wf_samples)
        # have to reshape to add an axis to broadcast `file_m` into `waveforms`
        waveforms[spk, :, :] = \
            file_m[spk_samples[0]:spk_samples[-1] + 1, ch].reshape((spk_samples.size, ch.size))
    print('Done. ({})'.format(time.ctime()))

    return waveforms



"""Extract stimulus timing info from ttl pulses and metadata."""

import os
import glob
import json
import numpy as np


def get_session_path(path):
    """
    Return local session path given a local path of any file from the session (assumes `Subjects`
    directory is present)

    Args:
        path (str): absolute path of any file in session
            e.g. /mnt/data/cortexlab/Subjects/CSHL050/2019-10-29/001/alf/spikes.times.npy
    Example usage:
        one = ONE()
        eid = one.search(subject=subject, date=date, number=number)
        files_paths = one.load(eid[0], download_only=True)
        session_path = get_session_path(files_paths)
    """
    from pathlib import Path
    subject_dir = np.where([part == 'Subjects' for part in Path(path).parts])[0]
    if len(subject_dir) == 0:
        raise FileNotFoundError('Could not find `Subject` directory in %s' % path)
    else:
        session_path = os.path.join(*Path(path).parts[:subject_dir[0] + 4])
    return session_path


def load_session_metadata(session_path):
    """

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :type session_path: str
    :return: dictionary of metadata
    :rtype: dict
    """
    json_file = glob.glob(os.path.join(session_path, 'raw_behavior_data', '*taskSettings.raw*'))[0]
    with open(json_file, 'r') as f:
        meta = json.load(f)
    return meta


def load_ttl_pulses(session_path):
    """
    Extract ttl pulses from sync signals

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :type session_path: str
    :return: ttl pulse times
    :rtype: np.ndarray
    """

    from ibllib.io.extractors.ephys_fpga import _get_main_probe_sync

    sync, sync_chmap = _get_main_probe_sync(session_path, bin_exists=False)
    fr2ttl_ch = sync_chmap['frame2ttl']

    # find times of when ttl polarity changes on fr2ttl channel
    sync_pol_ = sync['polarities'][sync['channels'] == fr2ttl_ch]
    sync_times_ = sync['times'][sync['channels'] == fr2ttl_ch]

    return sync_pol_, sync_times_


def _get_master_probe_dir(session_path):
    """Find master probe; code from ibllib.io.extractors.ephys_fpga._get_main_probe_sync"""
    from ibllib.io.extractors.ephys_fpga import _get_all_probes_sync
    from ibllib.io.extractors.ephys_fpga import get_neuropixel_version_from_files
    ephys_files = _get_all_probes_sync(session_path)
    if not ephys_files:
        raise FileNotFoundError(f"No ephys files found in {session_path}")
    version = get_neuropixel_version_from_files(ephys_files)
    if version == '3A':
        # the sync master is the probe with the most sync pulses
        sync_box_ind = np.argmax([ef.sync.times.size for ef in ephys_files])
    elif version == '3B':
        # the sync master is the nidq breakout box
        sync_box_ind = np.argmax([1 if ef.get('nidq') else 0 for ef in ephys_files])
    probe_dir = ephys_files[sync_box_ind].path
    return probe_dir


def load_rf_mapping_stimulus(session_path, stim_metadata):
    """
    extract frames of rf mapping stimulus

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :type session_path: str
    :param stim_metadata: dictionary of stimulus/task metadata
    :type stim_metadata: dict
    :return: stimulus frames
    :rtype: np.ndarray of shape (y_pix, x_pix, n_frames)
    """

    idx_rfm = get_stim_num_from_name(stim_metadata['VISUAL_STIMULI'], 'receptive_field_mapping')

    if idx_rfm is not None:
        stim_filename = stim_metadata['VISUAL_STIM_%i' % idx_rfm].get(
            'stim_data_file_name', '*RFMapStim.raw*')
        stim_file = glob.glob(os.path.join(session_path, 'raw_behavior_data', stim_filename))[0]
        frame_array = np.fromfile(stim_file, dtype='uint8')
        y_pix, x_pix, _ = stim_metadata['VISUAL_STIM_%i' % idx_rfm]['stim_file_shape']
        frames = np.transpose(np.reshape(frame_array, [y_pix, x_pix, -1], order='F'), [2, 1, 0])
    else:
        frames = np.array([])
    return frames


def get_stim_num_from_name(stim_ids, stim_name):
    """
    "VISUAL_STIMULI": {
    "0": "SPACER",
    "1": "receptive_field_mapping",
    "2": "orientation-direction_selectivity",
    "3": "contrast_reversal",
    "4": "task_stimuli",
    "5": "spontaneous_activity"}

    :param stim_ids: map from number (as string) to stimulus name
    :type stim_ids: dict
    :param stim_name: name of stimulus type
    :type stim_name: str
    :return: the number associated with the stimulus type
    :rtype: int
    """
    idx = None
    for key in stim_ids.keys():
        if stim_ids[key].lower() == stim_name.lower():
            idx = key
            break
    return int(idx)


def get_contrast_reversal_stimulus(stim_metadata):
    """
    Extract contrast reversal stimulus info

    :param stim_metadata: dictionary of stimulus metadata loaded from task json file
    :type stim_metadata: dict
    :return: np array of shape (n_stims,)
    :rtype: np.ndarray
    """
    stim_idx = get_stim_num_from_name(stim_metadata['VISUAL_STIMULI'], 'contrast_reversal')
    stim_metadata = stim_metadata['VISUAL_STIM_%i' % stim_idx]
    # return actual frames
    # patch_contrasts = stim_metadata['stim_patch_contrasts']
    # frames = np.stack([patch_contrasts[str(n)] for n in stim_metadata['stim_sequence']])
    # return index into actual frames
    stim = np.array(stim_metadata['stim_sequence'])
    return stim


def get_task_stimulus(session_path):
    """
    Extract contrast selectivity stimulus info

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :type session_path: str
    :return: np array of shape (n_stims, 2); col 0 contains stim azimuth, col 1 contains contrast
    :rtype: np.ndarray
    """
    from zipfile import ZipFile
    stim_file_zip = glob.glob(os.path.join(
        session_path, 'raw_behavior_data', '_iblrig_codeFiles*.zip'))[0]
    zf = ZipFile(stim_file_zip)
    # find task stimulus csv file
    stim_file = None
    for file in zf.namelist():
        if file.endswith('stims.csv'):
            stim_file = file
            break
    if stim_file is None:
        raise FileNotFoundError('Could not find a stims.csv file in the iblrig code files archive')
    with zf.open(stim_file) as f:
        stim_array = np.loadtxt(f, delimiter=' ')
    return stim_array


def get_rf_ttl_pulses(ttl_signal):
    """
    Find where ttl_signal increases or decreases

    :param ttl_signal:
    :type ttl_signal: array-like
    :return: where signal increases/decreases
    :rtype: tuple (np.ndarray, np.ndarray) of ttl (rise, fall) indices
    """
    # Convert values to 0, 1, -1 for simplicity
    assert len(np.unique(ttl_signal)) == 3
    ttl_sig = np.zeros(shape=ttl_signal.shape)
    ttl_sig[ttl_signal == np.max(ttl_signal)] = 1
    ttl_sig[ttl_signal == np.min(ttl_signal)] = -1
    # Find number of passage from 0->-1 and 0->+1
    d_ttl_sig = np.concatenate([np.diff(ttl_sig), [0]])
    idxs_up = np.where((ttl_sig == 0) & (d_ttl_sig == 1))[0]
    idxs_dn = np.where((ttl_sig == 0) & (d_ttl_sig == -1))[0]
    return idxs_up, idxs_dn


def get_expected_ttl_pulses(stim_order, stim_meta, ttl_signal_rf_map):
    """
    Get expected number of ttl pulses for each stimulus

    :param stim_order: list of stimulus ids throughout protocol
    :type stim_order: array-like
    :param stim_meta: dictionary containing stim metadata; from _iblrig_taskSettings json
    :type stim_meta: dict
    :param ttl_signal_rf_map: ttl signal during receptive field mapping with locally sparse noise
    :type ttl_signal_rf_map: array-like
    :return: list of ttl pulses for each stimulus class
    :rtype: list
    """
    n_expected_ttl_pulses = np.zeros(shape=(len(stim_order)))
    for i, stim_id in enumerate(stim_order):
        if stim_meta['VISUAL_STIMULI'][str(stim_id)] == 'receptive_field_mapping':
            n_instances = np.sum((np.array(stim_order) == stim_id) * 1)
            if n_instances > 1:
                raise ValueError('Extractor expects a single rf mapping presentation')
            # number of TTL pulses expected in frame2ttl trace for rf mapping
            idxs_up, idxs_dn = get_rf_ttl_pulses(ttl_signal_rf_map)
            n_expected_ttl_pulses[i] = len(idxs_up) + len(idxs_dn)
        else:
            key = str('VISUAL_STIM_%i' % stim_id)
            if key in stim_meta:
                n_expected_ttl_pulses[i] = stim_meta[key]['ttl_num']
            else:
                # spontaneous activity, no stimulus info in metadata
                n_expected_ttl_pulses[i] = 0
    return n_expected_ttl_pulses


# def update_ttl_pulses(stim_name, is_pol_beg_ok, is_pol_end_ok, stim_ts, n_expected_ttl_pulses):
#     """
#
#     :param stim_name: stim protocol name (e.g. 'receptive_field_mapping', 'contrast_reversal')
#     :type stim_name: str
#     :param is_pol_beg_ok:
#     :type is_pol_beg_ok: bool
#     :param is_pol_end_ok:
#     :type is_pol_end_ok: bool
#     :param stim_ts: stimulus presentation times
#     :type stim_ts: array-like
#     :param n_expected_ttl_pulses:
#     :type n_expected_ttl_pulses: int
#     :return: (new_stim_ts, new_expected_ttl_pulses)
#     :rtype: tuple
#     """
#
#     import copy
#     new_stim_ts = copy.copy(stim_ts)
#     new_expected_ttl_pulses = n_expected_ttl_pulses
#
#     if not is_pol_beg_ok:
#         if stim_name == 'receptive_field_mapping':
#             new_stim_ts = stim_ts[1:]
#             print('--> 1 TTL pulse removed from beginning')
#         elif stim_name == 'orientation-direction_selectivity':
#             new_stim_ts = stim_ts[1:]
#             print('--> 1 TTL pulse removed from beginning')
#         elif stim_name == 'contrast_reversal':
#             raise NotImplementedError
#         elif stim_name == 'task_stimuli':
#             new_stim_ts = stim_ts[1:]
#             print('--> 1 TTL pulse removed from beginning')
#         else:
#             raise ValueError('"%s" is not a valid stimulus protocol' % stim_name)
#
#     if not is_pol_end_ok:
#         if stim_name == 'receptive_field_mapping':
#             new_stim_ts = stim_ts[:-1]
#             print('--> 1 TTL pulse removed from end')
#         elif stim_name == 'orientation-direction_selectivity':
#             new_expected_ttl_pulses -= 1
#             print('--> updated correct number of TTL pulses')
#         elif stim_name == 'contrast_reversal':
#             raise NotImplementedError
#         elif stim_name == 'task_stimuli':
#             new_stim_ts = stim_ts[:-1]
#             print('--> 1 TTL pulse removed from end')
#         else:
#             raise ValueError('"%s" is not a valid stimulus protocol' % stim_name)
#
#     return new_stim_ts, new_expected_ttl_pulses


def get_spacer_times(spacer_template, jitter, ttl_signal, t_quiet):
    """
    :param spacer_template: list of indices where ttl signal changes
    :type spacer_template: array-like
    :param jitter: jitter (in seconds) for matching ttl_signal with spacer_template
    :type jitter: float
    :param ttl_signal:
    :type ttl_signal: array-like
    :param t_quiet: seconds between spacer and next stim
    :type t_quiet: float
    :return: times of spacer onset/offset
    :rtype: n_spacer x 2 np.ndarray; first col onset times, second col offset
    """
    diff_spacer_template = np.diff(spacer_template)
    # add jitter;
    # remove extreme values (shouldn't be a problem with iblrig versions >= 5.2.10)
    spacer_model = jitter + diff_spacer_template[2:-2]
    # diff ttl signal to compare to spacer_model
    dttl = np.diff(ttl_signal)
    # remove diffs larger than max diff in model to clean up signal
    dttl[dttl > np.max(spacer_model)] = 0
    # convolve cleaned diff ttl signal w/ spacer model
    conv_dttl = np.correlate(dttl, spacer_model, mode='full')
    # find spacer location
    thresh = 3.0
    idxs_spacer_middle = np.where(
        (conv_dttl[1:-2] < thresh) &
        (conv_dttl[2:-1] > thresh) &
        (conv_dttl[3:] < thresh))[0]
    # adjust indices for
    # - `np.where` call above
    # - length of spacer_model
    idxs_spacer_middle += 2 - int((np.floor(len(spacer_model) / 2)))
    # pull out spacer times (middle)
    ts_spacer_middle = ttl_signal[idxs_spacer_middle]
    # put beginning/end of spacer times into an array
    spacer_length = np.max(spacer_template)
    spacer_times = np.zeros(shape=(ts_spacer_middle.shape[0], 2))
    for i, t in enumerate(ts_spacer_middle):
        spacer_times[i, 0] = t - (spacer_length / 2) - t_quiet
        spacer_times[i, 1] = t + (spacer_length / 2) + t_quiet
    return spacer_times, conv_dttl


def interpolate_rf_mapping_stimulus(ttl_signal, times, frames, t_bin):
    """
    Interpolate stimulus presentation times to screen refresh rate to match `frames`

    :param ttl_signal:
    :type ttl_signal: array-like
    :param times: array of stimulus switch times
    :type times: array-like
    :param frames: (time, y_pix, x_pix) array of stim frames
    :type frames: array-like
    :param t_bin: screen refresh rate
    :type t_bin: float
    :return: tuple of (stim_times, stim_frames)
    """

    beg_extrap_val = -10001
    end_extrap_val = -10000

    idxs_up, idxs_dn = get_rf_ttl_pulses(ttl_signal)
    X = np.sort(np.concatenate([idxs_up, idxs_dn]))
    Xq = np.arange(frames.shape[0])
    # make left and right extrapolations distinctive to easily find later
    Tq = np.interp(Xq, X, times, left=beg_extrap_val, right=end_extrap_val)
    # uniform spacing outside boundaries of ttl signal
    # first values
    n_beg = len(np.where(Tq == beg_extrap_val)[0])
    if 0 < n_beg < Tq.shape[0]:
        Tq[:n_beg] = times[0] - np.arange(n_beg, 0, -1) * t_bin
    # end values
    n_end = len(np.where(Tq == end_extrap_val)[0])
    if 0 < n_end < Tq.shape[0]:
        Tq[-n_end:] = times[-1] + np.arange(1, n_end + 1) * t_bin
    return Tq, frames


def export_to_alf(session_path, stim_ts, stim_datas, stim_names):
    """
    Export extracted stimuli and their presentation times to the session alf directory.

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :param stim_ts:
    :param stim_datas:
    :param stim_names:
    :return: None; instead, saves the following files into `session_path/alf`; the 'xx' part of the
        following filenames will be '00', '01', etc., one for each time the specific protocol is
        run

        orientation/direction selectivity
        `_iblcertif_.odsgratings.times.xx.npy`: shape (n_stims, 2); columns are
            (stim on time, stim off time)
        `_iblcertif_.odsgratings.stims.xx.npy`: shape (n_stims,); value is grating angle

        contrast reversal
        `_iblcertif_.reversal.times.xx.npy`: shape (n_stims,); stim presentation times
        `_iblcertif_.reversal.stims.xx.npy`: shape (n_stims,); stim identity - these values map
            into the matrices defined in `_iblrig_taskSettings.raw.json` files, 'VISUAL_STIM_3' ->
            'stim_patch_contrasts'

        receptive field mapping (sparse noise)
        `_iblcertif_.rfmap.times.xx.npy`: shape (n_stims,); stim presentation times
        `_iblcertif_.rfmap.stims.xx.npy`: shape (n_stims, y_pix, x_pix); sparse noise stim frames

        spontaneous activity:
        `_iblcertif_.spontaneous.times.xx.npy`: shape (2,); start and end times of spont activity

        task stimuli (gratings of varying locations/contrast)
        `_iblcertif_.task.times.xx.npy`: shape (n_stims, 2); columns  are
            (stim on time, stim off time)
        `_iblcertif_.task.stims.xx.npy`: shape (n_stims, 2); columns are
            (azimuth in degrees [i.e. left or right], contrast)
    """

    counters = {stim_name: 0 for stim_name in np.unique(stim_names)}
    for stim_t, stim_data, stim_name in zip(stim_ts, stim_datas, stim_names):
        if stim_name == 'spacer':
            continue
        if stim_name == 'receptive_field_mapping':
            stim_short = 'rfmap'
        elif stim_name == 'orientation-direction_selectivity':
            stim_short = 'odsgratings'
        elif stim_name == 'contrast_reversal':
            stim_short = 'reversal'
        elif stim_name == 'task_stimuli':
            stim_short = 'task'
        elif stim_name == 'spontaneous_activity':
            stim_short = 'spontaneous'
        else:
            raise ValueError('"%s" is an unknown stimulus protocol' % stim_name)
        filename_times = str('_iblcertif_.%s.times.%02i.npy' % (stim_short, counters[stim_name]))
        filename_stims = str('_iblcertif_.%s.stims.%02i.npy' % (stim_short, counters[stim_name]))
        counters[stim_name] += 1
        if not os.path.exists(os.path.join(session_path, 'alf')):
            os.mkdir(os.path.join(session_path, 'alf'))
        shape_t = stim_t.shape[0]
        shape_d = stim_data.shape[0]
        if shape_t != 0 and shape_d != 0:
            if stim_name == 'contrast_reversal':
                # no known bug fix for contrast reversal bonsai bugs; don't save out if problem
                if shape_t != shape_d:
                    continue
            else:
                assert shape_t == shape_d
        if shape_t != 0:
            np.save(os.path.join(session_path, 'alf', filename_times), stim_t)
        if shape_d != 0:
            np.save(os.path.join(session_path, 'alf', filename_stims), stim_data)

    # alert cronjobs that there are new files to register to the database by saving empty file
    open(os.path.join(session_path, 'register_me.flag'), 'a').close()


def extract_stimulus_info_to_alf(session_path, t_bin=1 / 60, bin_jitter=3, save=True):
    """
    Extract the stimulus information stored in metadata and export to alf files. Also checks to
    make sure ttl pulses were extracted properly.
    Expected files/structure:
    - session metadata:
        session_path/raw_behavior_data/_iblrig_taskSettings.raw.*.json
    - task stimulus info:
        session_path/raw_behavior_data/_iblrig_codeFiles.raw.*.zip
    - rf mapping stim info:
        session_path/raw_behavior_data/_iblrig_RFMapStim.raw.*.bin
    - spikeglx sync data:
        session_path/raw_ephys_data/probe_[right/left]/_spikeglx_sync.[channels/polarities/times].*
            .npy

    :param session_path: absolute path of a session, i.e. /mnt/data/Subjects/ZM_1887/2019-07-10/001
    :type session_path: str
    :param t_bin: screen refresh rate
    :type t_bin: float
    :param bin_jitter: fudge factor in spacer template matching (units of time bins)
    :type bin_jitter: int
    :param save: export stimulus info to alf directory
    :type save: bool
    :return: None; instead stimulus info is stored in the following alf files:
            `_iblcertification_.stimtype.times.00.npy`
            `_iblcertification_.stimtype.frames.00.npy`
        where `stimtype` are `odsgratings`, `sparsenoise`, `contrastreversal`, and `taskstimulus`;
        see `export_to_alf` function documentation for more info on file structure
    """

    IBLRIG_VERSION_MIN = '5.2.9'

    # get ttl signal for extracting stim info (to compare with expected ttl signals in metadata)
    # ttl_sig = load_ttl_pulses(session_path)
    sync_pol, sync_times = load_ttl_pulses(session_path)

    # load session metadata
    meta = load_session_metadata(session_path)

    # basic checks
    protocol = meta['VISUAL_STIMULUS_TYPE']
    if protocol != 'ephys_certification':
        raise ValueError(
            '"%s" is an invalid protocol; function only supports "ephys_certification"' % protocol)
    iblrig_version = ''.join([i for i in meta['IBLRIG_VERSION_TAG'].split('.')])
    iblrig_version_min = ''.join([i for i in IBLRIG_VERSION_MIN.split('.')])
    if int(iblrig_version) < int(iblrig_version_min):
        raise ValueError(
            'Special extractor needed for code version {}; minimum supported version is {}'.format(
                iblrig_version, iblrig_version_min))
    print('extracting TTL pulses for iblrig version %s' % meta['IBLRIG_VERSION_TAG'])

    # pull useful fields out of metadata
    stim_ids = meta['VISUAL_STIMULI']
    stim_order = np.array(meta['STIM_ORDER'])
    frames = load_rf_mapping_stimulus(session_path, meta)
    id_spacer = get_stim_num_from_name(stim_ids, 'spacer')
    spacer_template = t_bin * np.array(meta['VISUAL_STIM_%i' % id_spacer]['ttl_frame_nums'])

    # get expected ttl pulses from upper left stim pixel (rf map) and metadata (all other stims)
    if len(frames) != 0:
        frame_ttl_signal = frames[:, 0, 0]
    else:
        frame_ttl_signal = None
    n_expected_ttl_pulses = get_expected_ttl_pulses(stim_order, meta, frame_ttl_signal)

    # get spacer info
    spacer_times, conv_sig = get_spacer_times(spacer_template, t_bin * bin_jitter, sync_times, 1)
    idxs_spacer = np.where(stim_order == get_stim_num_from_name(stim_ids, 'spacer'))[0]
    n_expected_spacers = len(idxs_spacer)
    n_spacers = spacer_times.shape[0]
    if n_spacers != n_expected_spacers:
        raise ValueError(
            '%i is an invalid number of spacer templates in ttl signal; expected %i' %
            (n_spacers, n_expected_spacers))
    else:
        print('found expected number of stimulus spacers')
    print('\n')

    # get stimulus info
    bonsai_jitter = 0.6  # bonsai timing can be slightly off
    stim_ts = [[] for _ in stim_order]
    stim_datas = [[] for _ in stim_order]
    stim_names = [[] for _ in stim_order]
    incorrect_pulses = 0
    for i, stim_id in enumerate(stim_order):
        stim_names[i] = stim_ids[str(stim_id)].lower()
        if i not in idxs_spacer:

            print('processing stimulus: %s' % stim_names[i])

            # assumes all non-spacers are preceded by a spacer
            sync_idxs = np.where(
                (sync_times > spacer_times[int(i / 2), 1]) &
                (sync_times < spacer_times[int((i + 1) / 2), 0]))[0]

            # ttl signal polarities are different depending on stimulus
            if stim_names[i] == 'receptive_field_mapping':
                pol_beg_expected = -1
                pol_end_expected = 1
            else:
                pol_beg_expected = 1
                pol_end_expected = -1
            # test beg/end polarities
            if stim_names[i] != 'spontaneous_activity':
                # check if initial polarity is correct
                if sync_pol[sync_idxs[0]] != pol_beg_expected:
                    is_pol_beg_ok = False
                    print('\twarning! wrong polarity detected at beginning of %s' % stim_names[i])
                else:
                    is_pol_beg_ok = True
                # check if final polarity is correct
                if sync_pol[sync_idxs[-1]] != pol_end_expected:
                    is_pol_end_ok = False
                    print('\twarning! wrong polarity detected at end of %s' % stim_names[i])
                else:
                    is_pol_end_ok = True
            else:
                is_pol_beg_ok = True
                is_pol_end_ok = True

            if stim_names[i] == 'receptive_field_mapping':
                stim_ts[i] = sync_times[sync_idxs]
                # account for polarity correctness
                if not is_pol_beg_ok:
                    stim_ts[i] = stim_ts[i][1:]
                    print('\t--> 1 TTL pulse removed from beginning')
                if not is_pol_end_ok:
                    stim_ts[i] = stim_ts[i][:-1]
                    print('\t--> 1 TTL pulse removed from end')
                stim_on_time = 0.2  # TODO: hardcoded for now
                # we only recorded rise times earlier; sync_times contains rise and fall times
                if np.diff(stim_ts[i][:2]) > stim_on_time / bonsai_jitter:
                    # get rid of bonsai artifacts at beginning; np.diff() will be too large
                    stim_ts[i] = stim_ts[i][2::2]
                    print('\tremoved bonsai onset artifact')
                else:
                    stim_ts[i] = stim_ts[i][0::2]
                stim_datas[i] = None  # handled below

            elif stim_names[i] == 'orientation-direction_selectivity':
                stim_ts[i] = sync_times[sync_idxs]
                # account for polarity correctness
                if not is_pol_beg_ok:
                    stim_ts[i] = stim_ts[i][1:]
                    print('\t--> 1 TTL pulse removed from beginning')
                if not is_pol_end_ok:
                    # stim_ts[i] = stim_ts[i][:-1]
                    n_expected_ttl_pulses[i] = n_expected_ttl_pulses[i] - 1  # -= 1 fails
                    print('\t--> 1 updated to correct number of TTL pulses')
                # separate ttl pulses for stim on and stim off
                # print(np.diff(stim_ts[i][:11]))
                # print(np.diff(stim_ts[i][-11:]))
                # print(stim_ts[i].shape)
                # print(len(meta['VISUAL_STIM_%i' % stim_id]['stim_sequence']))
                stim_on_time = meta['VISUAL_STIM_%i' % stim_id]['stim_on_time']
                if np.diff(stim_ts[i][:2]) < bonsai_jitter * stim_on_time:
                    # get rid of bonsai artifacts at beginning; np.diff() will be too small
                    stim_ts[i] = np.stack([stim_ts[i][2::2], stim_ts[i][3::2]], axis=1)
                    print('\tremoved bonsai onset artifact')
                else:
                    stim_ts[i] = np.stack([stim_ts[i][0::2], stim_ts[i][1::2]], axis=1)
                # ttl pulse for each stim rather than on/off
                n_expected_ttl_pulses[i] /= 2
                stim_sequence = meta['VISUAL_STIM_%i' % stim_id]['stim_sequence']
                stim_rads = meta['VISUAL_STIM_%i' % stim_id]['stim_directions_rad']
                # export direction of each grating in radians
                stim_datas[i] = np.array([stim_rads[str(j)] for j in stim_sequence])

            elif stim_names[i] == 'contrast_reversal':
                stim_ts[i] = sync_times[sync_idxs]
                # account for polarity correctness
                # if not (is_pol_beg_ok and is_pol_end_ok):
                #     raise NotImplementedError
                stim_datas[i] = get_contrast_reversal_stimulus(meta)

            elif stim_names[i] == 'task_stimuli':
                stim_ts[i] = sync_times[sync_idxs]
                # account for polarity correctness
                if not is_pol_beg_ok:
                    stim_ts[i] = stim_ts[i][1:]
                    print('\t--> 1 TTL pulse removed from beginning')
                if not is_pol_end_ok:
                    stim_ts[i] = stim_ts[i][:-1]
                    print('\t--> 1 TTL pulse removed from end')
                # separate ttl pulses for stim on and stim off
                stim_on_time = meta['VISUAL_STIM_%i' % stim_id]['stim_on_time']
                if np.diff(stim_ts[i][:2]) < bonsai_jitter * stim_on_time:
                    # get rid of bonsai artifacts at beginning; np.diff() will be too small
                    stim_ts[i] = np.stack([stim_ts[i][2::2], stim_ts[i][3::2]], axis=1)
                    print('\tremoved bonsai onset artifact')
                else:
                    stim_ts[i] = np.stack([stim_ts[i][0::2], stim_ts[i][1::2]], axis=1)
                # ttl pulse for each stim rather than on/off
                n_expected_ttl_pulses[i] /= 2
                stim_datas[i] = get_task_stimulus(session_path)

            elif stim_names[i] == 'spontaneous_activity':
                stim_ts[i] = np.array(
                    [spacer_times[int(i / 2), 1], spacer_times[int((i + 1) / 2), 0]])
                n_expected_ttl_pulses[i] = 2
                stim_datas[i] = np.array([])

            else:
                raise ValueError('"%s" is an unknown stimulus protocol' % stim_names[i])

            # check ttl pulses against expected ttl pulses from upper left stim pixel
            # (rf mapping) and metadata (all other stims)
            if stim_ts[i].shape[0] != n_expected_ttl_pulses[i]:
                incorrect_pulses += 1
                print(
                    '\t\033[1m\033[91mTTL pulses inconsistent for %s; expected %i, found %i\033[0m'
                    % (stim_names[i], n_expected_ttl_pulses[i], stim_ts[i].shape[0]))
                if stim_names[i] == 'contrast_reversal':
                    print(
                        '\twarning! no known fix for properly extracting contrast reversal ttl\n' +
                        '\tdo not trust exported contrast reversal info')
                    incorrect_pulses -= 1
            else:
                print('\tsuccessful stimulus extraction')
            print('\n')

    if incorrect_pulses == 0:
        print('found expected number of TTL pulses!')

    # assign proper timestamps for rf mapping
    if frame_ttl_signal is not None:
        # assumes there is only 1 presentation of RF mapping
        idx_rfs = np.where(
            stim_order == get_stim_num_from_name(stim_ids, 'receptive_field_mapping'))[0]
        if len(idx_rfs) > 1:
            raise NotImplementedError
        idx_rf = idx_rfs[0]
        stim_ts[idx_rf], stim_datas[idx_rf] = interpolate_rf_mapping_stimulus(
            frame_ttl_signal, stim_ts[idx_rf], frames, t_bin)

    if save and incorrect_pulses == 0:
        print('exporting stimulus information to %s' % os.path.join(session_path, 'alf/'))
        export_to_alf(session_path, stim_ts, stim_datas, stim_names)
    elif save and incorrect_pulses != 0:
        print('did not find expected TTL pulses; not saving extracted signals')


if __name__ == '__main__':

    # example usage
    from oneibl.one import ONE

    one = ONE()
    eid = one.search(subject='ZM_2104', date='2019-09-19', number=1)
    dtypes = [
        'ephysData.raw.meta',
        '_spikeglx_sync.channels',
        '_spikeglx_sync.polarities',
        '_spikeglx_sync.times',
        '_iblrig_RFMapStim.raw',
        '_iblrig_codeFiles.raw',
        '_iblrig_taskSettings.raw'
    ]
    files_paths = one.load(eid[0], dataset_types=dtypes, clobber=False, download_only=True)
    session_path = get_session_path(files_paths[0])
    extract_stimulus_info_to_alf(session_path, save=True)



import json
import shutil
from pathlib import Path

import alf.folders as folders
from alf.io import is_uuid_string
import ibllib.io.flags as flags
import ibllib.io.params as params
import ibllib.io.spikeglx as spikeglx
from oneibl.one import ONE


def cli_ask_default(prompt: str, default: str):
    dflt = " [default: {}]: "
    dflt = dflt.format(default)
    ans = input(prompt + dflt) or default
    return ans


def cli_ask_options(prompt: str, options: list, default_idx: int = 0) -> str:
    parsed_options = [str(x) for x in options]
    if default_idx is not None:
        parsed_options[default_idx] = f"[{parsed_options[default_idx]}]"
    options_str = ' (' + ' | '.join(parsed_options) + ')> '
    ans = input(prompt + options_str) or str(options[default_idx])
    if ans not in [str(x) for x in options]:
        return cli_ask_options(prompt, options, default_idx=default_idx)
    return ans


def behavior_exists(session_path: str) -> bool:  # testOK
    session_path = Path(session_path)
    behavior_path = session_path / 'raw_behavior_data'
    if behavior_path.exists():
        return True
    return False


def check_transfer(src_session_path: str, dst_session_path: str):
    # missing fail test
    src_files = sorted([x for x in Path(src_session_path).rglob('*') if x.is_file()])
    dst_files = sorted([x for x in Path(dst_session_path).rglob('*') if x.is_file()])
    assert_msg = 'src_files != dst_files'
    assert len(src_files) == len(dst_files), assert_msg
    for s, d in zip(src_files, dst_files):
        assert s.name == d.name, assert_msg
        assert s.stat().st_size == d.stat().st_size, assert_msg


def rename_session(session_path: str) -> Path:
    session_path = Path(folders.session_path(session_path))
    if session_path is None:
        return
    mouse = session_path.parts[-3]
    date = session_path.parts[-2]
    sess = session_path.parts[-1]
    new_mouse = input(
        f"Please insert mouse NAME [current value: {mouse}]> ") or mouse
    new_date = input(
        f"Please insert new session DATE [current value: {date}]> ") or date
    new_sess = input(
        f"Please insert new session NUMBER [current value: {sess}]> ") or sess
    new_session_path = Path(*session_path.parts[:-3]) / new_mouse / new_date / new_sess

    shutil.move(str(session_path), str(new_session_path))
    print(session_path, '--> renamed to:')
    print(new_session_path)

    return new_session_path


def transfer_folder(src: Path, dst: Path, force: bool = False) -> None:
    print(f"Attempting to copy:\n{src}\n--> {dst}")
    if force:
        print(f"Removing {dst}")
        shutil.rmtree(dst, ignore_errors=True)
    print(f"Copying all files:\n{src}\n--> {dst}")
    shutil.copytree(src, dst)
    # If folder was created delete the src_flag_file
    if check_transfer(src, dst) is None:
        print("All files copied")


def load_params_dict(params_fname: str) -> dict:
    params_fpath = Path(params.getfile(params_fname))
    if not params_fpath.exists():
        return None
    with open(params_fpath, 'r') as f:
        out = json.load(f)
    return out


def load_videopc_params():
    if not load_params_dict('videopc_params'):
        create_videopc_params()
    return load_params_dict('videopc_params')


def load_ephyspc_params():
    if not load_params_dict('ephyspc_params'):
        create_ephyspc_params()
    return load_params_dict('ephyspc_params')


def create_videopc_params(force=False, silent=False):
    if Path(params.getfile('videopc_params')).exists() and not force:
        print(f"{params.getfile('videopc_params')} exists already, exiting...")
        print(Path(params.getfile('videopc_params')).exists())
        return
    if silent:
        data_folder_path = r"D:\iblrig_data\Subjects"
        remote_data_folder_path = r"\\iblserver.champalimaud.pt\ibldata\Subjects"
        body_cam_idx = 0
        left_cam_idx = 1
        right_cam_idx = 2
    else:
        data_folder_path = cli_ask_default(
            r"Where's your LOCAL 'Subjects' data folder?", r"D:\iblrig_data\Subjects")
        remote_data_folder_path = cli_ask_default(
            r"Where's your REMOTE 'Subjects' data folder?",
            r"\\iblserver.champalimaud.pt\ibldata\Subjects")
        body_cam_idx = cli_ask_default("Please select the index of the BODY camera", '0')
        left_cam_idx = cli_ask_default("Please select the index of the LEFT camera", '1')
        right_cam_idx = cli_ask_default("Please select the index of the RIGHT camera", '2')

    param_dict = {
        'DATA_FOLDER_PATH': data_folder_path,
        'REMOTE_DATA_FOLDER_PATH': remote_data_folder_path,
        'BODY_CAM_IDX': body_cam_idx,
        'LEFT_CAM_IDX': left_cam_idx,
        'RIGHT_CAM_IDX': right_cam_idx,
    }
    params.write('videopc_params', param_dict)
    print(f"Created {params.getfile('videopc_params')}")
    print(param_dict)
    return param_dict


def create_ephyspc_params(force=False, silent=False):
    if Path(params.getfile('ephyspc_params')).exists() and not force:
        print(f"{params.getfile('ephyspc_params')} exists already, exiting...")
        print(Path(params.getfile('ephyspc_params')).exists())
        return
    if silent:
        data_folder_path = r"D:\iblrig_data\Subjects"
        remote_data_folder_path = r"\\iblserver.champalimaud.pt\ibldata\Subjects"
        probe_type_00 = '3A'
        probe_type_01 = '3B'
    else:
        data_folder_path = cli_ask_default(
            r"Where's your LOCAL 'Subjects' data folder?", r"D:\iblrig_data\Subjects")
        remote_data_folder_path = cli_ask_default(
            r"Where's your REMOTE 'Subjects' data folder?",
            r"\\iblserver.champalimaud.pt\ibldata\Subjects")
        probe_type_00 = cli_ask_options("What's the type of PROBE 00?", ['3A', '3B'])
        probe_type_01 = cli_ask_options("What's the type of PROBE 01?", ['3A', '3B'])
    param_dict = {
        'DATA_FOLDER_PATH': data_folder_path,
        'REMOTE_DATA_FOLDER_PATH': remote_data_folder_path,
        'PROBE_TYPE_00': probe_type_00,
        'PROBE_TYPE_01': probe_type_01,
    }
    params.write('ephyspc_params', param_dict)
    print(f"Created {params.getfile('ephyspc_params')}")
    print(param_dict)
    return param_dict


def confirm_video_remote_folder(local_folder=False, remote_folder=False, force=False):
    pars = load_videopc_params()

    if not local_folder:
        local_folder = pars['DATA_FOLDER_PATH']
    if not remote_folder:
        remote_folder = pars['REMOTE_DATA_FOLDER_PATH']
    local_folder = Path(local_folder)
    remote_folder = Path(remote_folder)
    # Check for Subjects folder
    local_folder = folders.subjects_data_folder(local_folder, rglob=True)
    remote_folder = folders.subjects_data_folder(remote_folder, rglob=True)

    print('LOCAL:', local_folder)
    print('REMOTE:', remote_folder)
    src_session_paths = [x.parent for x in local_folder.rglob("transfer_me.flag")]

    if not src_session_paths:
        print("Nothing to transfer, exiting...")
        return

    for session_path in src_session_paths:
        print(f"\nFound session: {session_path}")
        flag_file = session_path / 'transfer_me.flag'
        msg = f"Transfer to {remote_folder} with the same name?"
        resp = input(msg + "\n[y]es/[r]ename/[s]kip/[e]xit\n ^\n> ") or 'y'
        resp = resp.lower()
        print(resp)
        if resp not in ['y', 'r', 's', 'e', 'yes', 'rename', 'skip', 'exit']:
            return confirm_video_remote_folder(
                local_folder=local_folder, remote_folder=remote_folder, force=force)
        elif resp == 'y' or resp == 'yes':
            pass
        elif resp == 'r' or resp == 'rename':
            session_path = rename_session(session_path)
        elif resp == 's' or resp == 'skip':
            continue
        elif resp == 'e' or resp == 'exit':
            return

        remote_session_path = remote_folder / Path(*session_path.parts[-3:])
        if not behavior_exists(remote_session_path):
            print(f"No behavior folder found in {remote_session_path}: skipping session...")
            return
        transfer_folder(
            session_path / 'raw_video_data',
            remote_session_path / 'raw_video_data',
            force=force)
        flag_file.unlink()
        create_video_transfer_done_flag(remote_session_path)
        check_create_raw_session_flag(remote_session_path)


def confirm_ephys_remote_folder(local_folder=False, remote_folder=False,
                                force=False, iblscripts_folder=False):
    pars = load_ephyspc_params()

    if not local_folder:
        local_folder = pars['DATA_FOLDER_PATH']
    if not remote_folder:
        remote_folder = pars['REMOTE_DATA_FOLDER_PATH']
    local_folder = Path(local_folder)
    remote_folder = Path(remote_folder)
    # Check for Subjects folder
    local_folder = folders.subjects_data_folder(local_folder, rglob=True)
    remote_folder = folders.subjects_data_folder(remote_folder, rglob=True)

    print('LOCAL:', local_folder)
    print('REMOTE:', remote_folder)
    src_session_paths = [x.parent for x in local_folder.rglob("transfer_me.flag")]

    if not src_session_paths:
        print("Nothing to transfer, exiting...")
        return

    for session_path in src_session_paths:
        print(f"\nFound session: {session_path}")
        # Rename ephys files
        # FIXME: if transfer has failed and wiring file is there renaming will fail!
        rename_ephys_files(str(session_path))
        # Move ephys files
        move_ephys_files(str(session_path))
        # Copy wiring files
        copy_wiring_files(str(session_path), iblscripts_folder)
        try:
            create_alyx_probe_insertions(str(session_path))
        except BaseException as e:
            print(e, "\nCreation failed, please create the probe insertions manually.",
                  "Continuing transfer...")
        msg = f"Transfer to {remote_folder} with the same name?"
        resp = input(msg + "\n[y]es/[r]ename/[s]kip/[e]xit\n ^\n> ") or 'y'
        resp = resp.lower()
        print(resp)
        if resp not in ['y', 'r', 's', 'e', 'yes', 'rename', 'skip', 'exit']:
            return confirm_ephys_remote_folder(
                local_folder=local_folder, remote_folder=remote_folder,
                force=force, iblscripts_folder=iblscripts_folder)
        elif resp == 'y' or resp == 'yes':
            pass
        elif resp == 'r' or resp == 'rename':
            session_path = rename_session(session_path)
        elif resp == 's' or resp == 'skip':
            continue
        elif resp == 'e' or resp == 'exit':
            return

        remote_session_path = remote_folder / Path(*session_path.parts[-3:])
        if not behavior_exists(remote_session_path):
            print(f"No behavior folder found in {remote_session_path}: skipping session...")
            return
        # TODO: Check flagfiles on src.and dst + alf dir in session folder then remove
        # Try catch? wher catch condition is force transfer maybe
        transfer_folder(
            session_path / 'raw_ephys_data',
            remote_session_path / 'raw_ephys_data',
            force=force)
        # if behavior extract_me.flag exists remove it, because of ephys flag
        flag_file = session_path / 'transfer_me.flag'
        flag_file.unlink()
        if (remote_session_path / 'extract_me.flag').exists():
            (remote_session_path / 'extract_me.flag').unlink()
        # Create remote flags
        create_ephys_transfer_done_flag(remote_session_path)
        check_create_raw_session_flag(remote_session_path)


def create_alyx_probe_insertions(session_path: str,
                                 force: bool = False,
                                 one: object = None,
                                 model: str = None,
                                 labels: list = None):
    if one is None:
        one = ONE()
    if is_uuid_string(session_path):
        eid = session_path
    else:
        eid = one.eid_from_path(session_path)
    if eid is None:
        print('Session not found on Alyx: please create session before creating insertions')
    if model is None:
        probe_model = spikeglx.get_neuropixel_version_from_folder(session_path)
        pmodel = '3B2' if probe_model == '3B' else probe_model
    else:
        pmodel = model
    raw_ephys_data_path = Path(session_path) / 'raw_ephys_data'
    if labels is None:
        probe_labels = [
            x.name for x in Path(raw_ephys_data_path).glob('*')
            if x.is_dir() and ('00' in x.name or '01' in x.name)
        ]
    else:
        probe_labels = labels
    # create the dictionary
    for plabel in probe_labels:
        insdict = {'session': eid,
                   'name': plabel,
                   'model': pmodel}
        # search for the corresponding insertion in Alyx
        alyx_insertion = one.alyx.rest('insertions', 'list',
                                       session=insdict['session'],
                                       name=insdict['name'])
        # if it doesn't exist, create it
        if len(alyx_insertion) == 0:
            alyx_insertion = one.alyx.rest('insertions', 'create', data=insdict)
        else:
            iid = alyx_insertion[0]['id']
            if force:
                alyx_insertion = one.alyx.rest('insertions', 'update', id=iid, data=insdict)
            else:
                alyx_insertion = alyx_insertion[0]


def create_ephys_flags(session_folder: str):
    session_path = Path(session_folder)
    flags.write_flag_file(session_path.joinpath('extract_ephys.flag'))
    flags.write_flag_file(session_path.joinpath('raw_ephys_qc.flag'))
    probe00_path = session_path / 'raw_ephys_data' / 'probe00'
    probe01_path = session_path / 'raw_ephys_data' / 'probe01'
    flags.write_flag_file(probe00_path.joinpath('spike_sorting.flag'))
    flags.write_flag_file(probe01_path.joinpath('spike_sorting.flag'))


def create_ephys_transfer_done_flag(session_folder: str) -> None:
    session_path = Path(session_folder)
    flags.write_flag_file(session_path.joinpath('ephys_data_transferred.flag'))


def create_video_transfer_done_flag(session_folder: str) -> None:
    session_path = Path(session_folder)
    flags.write_flag_file(session_path.joinpath('video_data_transferred.flag'))


def check_create_raw_session_flag(session_folder: str) -> None:
    session_path = Path(session_folder)
    ephys = session_path.joinpath('ephys_data_transferred.flag')
    video = session_path.joinpath('video_data_transferred.flag')
    if ephys.exists() and video.exists():
        flags.write_flag_file(session_path.joinpath('raw_session.flag'))
        ephys.unlink()
        video.unlink()


def rename_ephys_files(session_folder: str) -> None:
    """rename_ephys_files is system agnostic (3A, 3B1, 3B2).
    Renames all ephys files to Alyx compatible filenames. Uses get_new_filename.

    :param session_folder: Session folder path
    :type session_folder: str
    :return: None - Changes names of files on filesystem
    :rtype: None
    """
    session_path = Path(session_folder)
    ap_files = session_path.rglob('*.ap.*')
    lf_files = session_path.rglob('*.lf.*')
    nidq_files = session_path.rglob('*.nidq.*')

    for apf in ap_files:
        new_filename = get_new_filename(apf.name)
        shutil.move(str(apf), str(apf.parent / new_filename))

    for lff in lf_files:
        new_filename = get_new_filename(lff.name)
        shutil.move(str(lff), str(lff.parent / new_filename))

    for nidqf in nidq_files:
        new_filename = get_new_filename(nidqf.name)
        shutil.move(str(nidqf), str(nidqf.parent / new_filename))


def get_new_filename(filename: str) -> str:  # testOK
    """get_new_filename is system agnostic (3A, 3B1, 3B2).
    Gets an alyx compatible filename from any spikeglx ephys file.

    :param filename: Name of an ephys file
    :type filename: str
    :return: New name for ephys file
    :rtype: str
    """
    root = '_spikeglx_ephysData'
    # gt = '_g0_t0' or 'error
    gt = None
    for g in range(10):
        for t in range(10):
            if f'_g{g}_t{t}' in filename:
                gt = f'_g{g}_t{t}'
    else:
        if gt is None:
            raise NotImplementedError(f"Can not parse {filename} as an ephys file with gX_tY where"
                                      f"XY are single digits")

    # ext = 'bin' or 'meta'
    if '.bin' in filename:
        ext = 'bin'
    elif '.meta' in filename:
        ext = 'meta'

    if '.nidq.' in filename:
        return '.'.join([root + gt, 'nidq', ext])

    # probe = 'imec0' or 'imec1'
    if '.imec0.' in filename:
        probe = 'imec0'
    elif '.imec1.' in filename:
        probe = 'imec1'
    elif '.imec.' in filename:
        probe = 'imec'

    # freq = 'ap' or 'lf'
    if '.ap.' in filename:
        freq = 'ap'
    elif '.lf.' in filename:
        freq = 'lf'

    return '.'.join([root + gt, probe, freq, ext])


def move_ephys_files(session_folder: str) -> None:
    """move_ephys_files is system agnostic (3A, 3B1, 3B2).
    Moves all properly named ephys files to appropriate locations for transfer.
    Use rename_ephys_files function before this one.

    :param session_folder: Session folder path
    :type session_folder: str
    :return: None - Moves files on filesystem
    :rtype: None
    """
    session_path = Path(session_folder)
    raw_ephys_data_path = session_path / 'raw_ephys_data'

    probe00_path = raw_ephys_data_path / 'probe00'
    probe01_path = raw_ephys_data_path / 'probe01'
    # 3A system imec only
    imec_files = session_path.rglob('*.imec.*')
    for imf in imec_files:
        if 'probe00' in str(imf):
            shutil.move(str(imf), str(probe00_path / imf.name))
        elif 'probe01' in str(imf):
            shutil.move(str(imf), str(probe01_path / imf.name))
    # 3B system
    imec0_files = session_path.rglob('*.imec0.*')
    imec1_files = session_path.rglob('*.imec1.*')
    nidq_files = session_path.rglob('*.nidq.*')
    # All imec 0 in probe00 folder
    for i0f in imec0_files:
        shutil.move(str(i0f), str(probe00_path / i0f.name))
    # All imec 1 in probe01 folder
    for i1f in imec1_files:
        shutil.move(str(i1f), str(probe01_path / i1f.name))
    # NIDAq files
    nidq_files = session_path.rglob('*.nidq.*')
    for nidqf in nidq_files:
        shutil.move(str(nidqf), str(raw_ephys_data_path / nidqf.name))
    # Delete all empty folders recursively
    from ibllib.io.misc import delete_empty_folders
    delete_empty_folders(raw_ephys_data_path, dry=False, recursive=True)


def create_custom_ephys_wirings(iblscripts_folder: str):
    iblscripts_path = Path(iblscripts_folder)
    PARAMS = load_ephyspc_params()
    if PARAMS['PROBE_TYPE_00'] != PARAMS['PROBE_TYPE_01']:
        print("Having different probe types not supported")
        raise(NotImplementedError)

    params_path = iblscripts_path.parent / 'iblscripts_params'
    params_path.mkdir(parents=True, exist_ok=True)
    wirings_path = iblscripts_path / 'deploy' / 'ephyspc' / 'wirings'
    if '3A' in PARAMS['PROBE_TYPE_00']:
        shutil.copy(str(wirings_path / '3A.wiring.json'),
                    str(params_path / '3A_probe00.wiring.json'))
        shutil.copy(str(wirings_path / '3A.wiring.json'),
                    str(params_path / '3A_probe01.wiring.json'))
        print(f'Created 3A.wiring.json in {params_path}')
    if '3B' in PARAMS['PROBE_TYPE_00']:
        shutil.copy(str(wirings_path / '3B.wiring.json'),
                    str(params_path / '3B_probe00.wiring.json'))
        shutil.copy(str(wirings_path / '3B.wiring.json'),
                    str(params_path / '3B_probe01.wiring.json'))
        print(f'Created 3B.wiring.json in {params_path}')
        shutil.copy(str(wirings_path / 'nidq.wiring.json'),
                    str(params_path / 'nidq.wiring.json'))
        print(f'Created nidq.wiring.json in {params_path}')
    print(f'\nYou can now modify your wirings files from folder {params_path}')


def get_iblscripts_folder():
    return str(Path().cwd().parent.parent)


def copy_wiring_files(session_folder, iblscripts_folder):
    """Run after moving files to probe folders"""
    PARAMS = load_ephyspc_params()
    if PARAMS['PROBE_TYPE_00'] != PARAMS['PROBE_TYPE_01']:
        print("Having different probe types is not supported")
        raise(NotImplementedError)
    session_path = Path(session_folder)
    iblscripts_path = Path(iblscripts_folder)
    iblscripts_params_path = iblscripts_path.parent / 'iblscripts_params'
    wirings_path = iblscripts_path / 'deploy' / 'ephyspc' / 'wirings'
    probe00_path = session_path / 'raw_ephys_data' / 'probe00'
    probe01_path = session_path / 'raw_ephys_data' / 'probe01'
    termination = '.wiring.json'
    # Determine if custom wiring applies
    custom = True if iblscripts_params_path.exists() else False
    # Determine system
    ephys_system = PARAMS['PROBE_TYPE_00']
    # Define where to get the files from
    if custom:
        src_wiring_path = iblscripts_params_path
        probe00_wiring_file_path = src_wiring_path / f'{ephys_system}_probe00.wiring.json'
        probe01_wiring_file_path = src_wiring_path / f'{ephys_system}_probe01.wiring.json'
    else:
        src_wiring_path = wirings_path
        probe00_wiring_file_path = src_wiring_path / f'{ephys_system}.wiring.json'
        probe01_wiring_file_path = src_wiring_path / f'{ephys_system}.wiring.json'

    if ephys_system == '3B':
        # Copy nidq file
        nidq_files = session_path.rglob('*.nidq.bin')
        for nidqf in nidq_files:
            nidq_wiring_name = '.'.join(str(nidqf.name).split('.')[:-1]) + termination
            shutil.copy(str(src_wiring_path / 'nidq.wiring.json'),
                        str(session_path / 'raw_ephys_data' / nidq_wiring_name))
    # If system is either (3A OR 3B) copy a wiring file for each ap.bin file
    for binf in session_path.rglob('*.ap.bin'):
        wiring_name = '.'.join(str(binf.name).split('.')[:-2]) + termination
        if 'probe00' in str(binf):
            shutil.copy(str(probe00_wiring_file_path),
                        str(probe00_path / wiring_name))
        if 'probe01' in str(binf):
            shutil.copy(str(probe01_wiring_file_path),
                        str(probe01_path / wiring_name))



from pathlib import Path
import abc
import logging
import io
import importlib
import time
from _collections import OrderedDict
import traceback

from graphviz import Digraph

from ibllib.misc import version
from ibllib.io import params
from oneibl.registration import register_dataset


_logger = logging.getLogger('ibllib')


class Task(abc.ABC):
    log = ""
    cpu = 1
    gpu = 0
    io_charge = 5  # integer percentage
    priority = 30  # integer percentage, 100 means highest priority
    ram = 4  # RAM needed to run (Go)
    one = None  # one instance (optional)
    level = 0
    outputs = None
    time_elapsed_secs = None
    time_out_secs = None
    version = version.ibllib()

    def __init__(self, session_path, parents=None, taskid=None, one=None):
        self.taskid = taskid
        self.one = one
        self.session_path = session_path
        self.register_kwargs = {}
        if parents:
            self.parents = parents
        else:
            self.parents = []

    @property
    def name(self):
        return self.__class__.__name__

    def run(self, **kwargs):
        """
        --- do not overload, see _run() below---
        wraps the _run() method with
        -   error management
        -   logging to variable
        """
        # if taskid of one properties are not available, local run only without alyx
        use_alyx = self.one is not None and self.taskid is not None
        if use_alyx:
            self.one.alyx.rest('tasks', 'partial_update',
                               id=self.taskid, data={'status': 'Started'})
        # setup
        self.setUp()
        # Setup the console handler with a StringIO object
        log_capture_string = io.StringIO()
        ch = logging.StreamHandler(log_capture_string)
        str_format = '%(asctime)s,%(msecs)d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s'
        ch.setFormatter(logging.Formatter(str_format))
        _logger.addHandler(ch)
        _logger.info(f"Starting job {self.__class__}")
        # run
        start_time = time.time()
        try:
            self.outputs = self._run(**kwargs)
            self.status = 0
            _logger.info(f"Job {self.__class__} complete")
        except BaseException:
            _logger.error(traceback.format_exc())
            _logger.info(f"Job {self.__class__} errored")
            self.status = -1
        self.time_elapsed_secs = time.time() - start_time
        # log the outputs-+
        if isinstance(self.outputs, list):
            nout = len(self.outputs)
        elif self.outputs is None:
            nout = 0
        else:
            nout = 1
        _logger.info(f"N outputs: {nout}")
        _logger.info(f"--- {self.time_elapsed_secs} seconds run-time ---")
        # after the run, capture the log output
        self.log = log_capture_string.getvalue()
        log_capture_string.close()
        _logger.removeHandler(ch)
        # tear down
        self.tearDown()
        return self.status

    def register_datasets(self, one=None, **kwargs):
        """
        Register output datasets form the task to Alyx
        :param one:
        :param jobid:
        :param kwargs: directly passed to the register_dataset function
        :return:
        """
        assert one
        if self.outputs:
            if isinstance(self.outputs, list):
                versions = [self.version for _ in self.outputs]
            else:
                versions = [self.version]
            return register_dataset(self.outputs, one=one, versions=versions, **kwargs)

    def rerun(self):
        self.run(overwrite=True)

    @abc.abstractmethod
    def _run(self, overwrite=False):
        """
        This is the method to implement
        :param overwrite: (bool) if the output already exists,
        :return: out_files: files to be registered. Could be a list of files (pathlib.Path),
        a single file (pathlib.Path) an empty list [] or None.
        Whithin the pipeline, there is a distinction between a job that returns an empty list
         and a job that returns None. If the function returns None, the job will be labeled as
          "empty" status in the database, otherwise, the job has an expected behaviour of not
          returning any dataset.
        """

    def setUp(self):
        """
        Function to optionally overload to check inputs.
        :return:
        """

    def tearDown(self):
        """
        Function to optionally overload to check results
        :return:
        """


class Pipeline(abc.ABC):
    """
    Pipeline class: collection of related and potentially interdependent tasks
    """
    tasks = OrderedDict()
    one = None

    def __init__(self, session_path=None, one=None, eid=None):
        assert session_path or eid
        self.one = one
        self.eid = eid
        if session_path:
            self.session_path = session_path
            self.eid = one.eid_from_path(session_path) if self.one else None
        self.label = self.__module__ + '.' + type(self).__name__

    def make_graph(self, out_dir=None, show=True):
        if not out_dir:
            par = params.read('one_params')
            out_dir = par.CACHE_DIR
        m = Digraph('G', filename=str(Path(out_dir).joinpath(self.__module__ + '_graphs.gv')))
        m.attr(rankdir='TD')

        e = Digraph(name='cluster_' + self.label)
        e.attr('node', shape='box')
        e.node('root', label=self.label)

        e.attr('node', shape='ellipse')
        for k in self.tasks:
            j = self.tasks[k]
            if len(j.parents) == 0:
                e.edge('root', j.name)
            else:
                [e.edge(p.name, j.name) for p in j.parents]

        m.subgraph(e)
        m.attr(label=r'\n\Pre-processing\n')
        m.attr(fontsize='20')
        if show:
            m.view()
        return m

    def create_alyx_tasks(self, rerun__status__in=None):
        """
        Instantiate the pipeline and create the tasks in Alyx, then create the jobs for the session
        If the jobs already exist, they are left untouched. The re-run parameter will re-init the
        job by emptying the log and set the status to Waiting
        :param rerun__status__in: by default no re-run. To re-run tasks if they already exist,
        specify a list of statuses string that will be re-run, those are the possible choices:
        ['Waiting', 'Started', 'Errored', 'Empty', 'Complete']
        to always patch, the string '__all__' can also be provided
        :return: list of alyx tasks dictionaries (existing and or created)
        """
        rerun__status__in = rerun__status__in or []
        if rerun__status__in == '__all__':
            rerun__status__in = ['Waiting', 'Started', 'Errored', 'Empty', 'Complete']
        assert self.eid
        if self.one is None:
            _logger.warning("No ONE instance found for Alyx connection, set the one property")
            return
        tasks_alyx_pre = self.one.alyx.rest('tasks', 'list', session=self.eid, graph=self.name)
        tasks_alyx = []
        # creates all the tasks by iterating through the ordered dict
        for k, t in self.tasks.items():
            # get the parents alyx ids to reference in the database
            if len(t.parents):
                pnames = [p.name for p in t.parents]
                parents_ids = [ta['id'] for ta in tasks_alyx if ta['name'] in pnames]
            else:
                parents_ids = []
            task_dict = {'executable': f"{t.__module__}.{t.name}", 'priority': t.priority,
                         'io_charge': t.io_charge, 'gpu': t.gpu, 'cpu': t.cpu,
                         'ram': t.ram, 'module': self.label, 'parents': parents_ids,
                         'level': t.level, 'time_out_sec': t.time_out_secs, 'session': self.eid,
                         'status': 'Waiting', 'log': None, 'name': t.name, 'graph': self.name}
            # if the task already exists, patch it otherwise, create it
            talyx = next(filter(lambda x: x["name"] == t.name, tasks_alyx_pre), [])
            if len(talyx) == 0:
                talyx = self.one.alyx.rest('tasks', 'create', data=task_dict)
            elif rerun__status__in == '__all__' or talyx['status'] in rerun__status__in:
                talyx = self.one.alyx.rest(
                    'tasks', 'partial_update', id=talyx['id'], data=task_dict)
            tasks_alyx.append(talyx)
        return tasks_alyx

    def run(self, status__in=['Waiting'], **kwargs):
        """
        Get all the session related jobs from alyx and run them
        :param status__in: lists of status strings to run in
        ['Waiting', 'Started', 'Errored', 'Empty', 'Complete']
        :param kwargs: arguments passed downstream to run_alyx_task
        :return: jalyx: list of REST dictionaries of the job endpoints
        :return: job_deck: list of REST dictionaries of the jobs endpoints
        :return: all_datasets: list of REST dictionaries of the dataset endpoints
        """
        assert self.session_path, "Pipeline object has to be declared with a session path to run"
        if self.one is None:
            _logger.warning("No ONE instance found for Alyx connection, set the one property")
            return
        task_deck = self.one.alyx.rest('tasks', 'list', session=self.eid)
        # [(t['name'], t['level']) for t in task_deck]
        all_datasets = []
        for i, j in enumerate(task_deck):
            if j['status'] not in status__in:
                continue
            # here we update the status in-place to avoid another hit to the database
            task_deck[i], dsets = run_alyx_task(tdict=j, session_path=self.session_path,
                                                one=self.one, job_deck=task_deck, **kwargs)
            if dsets is not None:
                all_datasets.extend(dsets)
        return task_deck, all_datasets

    def rerun_failed(self):
        return self.run(status__in=['Waiting', 'Held', 'Started', 'Errored', 'Empty'])

    def rerun(self):
        return self.run(status__in=['Waiting', 'Held', 'Started', 'Errored', 'Empty', 'Complete'])

    @property
    def name(self):
        return self.__class__.__name__


def run_alyx_task(tdict=None, session_path=None, one=None, job_deck=None, max_md5_size=None):
    """
    Runs a single Alyx job and registers output datasets
    :param tdict:
    :param session_path:
    :param one:
    :param job_deck: optional list of job dictionaries belonging to the session. Needed
    to check dependency status if the jdict has a parent field. If jdict has a parent and
    job_deck is not entered, will query the database
    :param max_md5_size: in bytes, if specified, will not compute the md5 checksum above a given
    filesize to save time
    :return:
    """
    registered_dsets = []
    if len(tdict['parents']):
        # here we need to check parents status, get the job_deck if not available
        if not job_deck:
            job_deck = one.alyx.rest('tasks', 'list', session=tdict['session'])
        # check the dependencies
        parent_tasks = filter(lambda x: x['id'] in tdict['parents'], job_deck)
        parent_statuses = [j['status'] for j in parent_tasks]
        # if any of the parent tasks is not complete, throw a warning
        if any(map(lambda s: s != 'Complete', parent_statuses)):
            _logger.warning(f"{tdict['name']} has unmet dependencies")
            # if parents are just waiting, don't do anything, but if they have a failed status
            # set the current task status to Held
            if any(map(lambda s: s in ['Errored', 'Held', 'Empty'], parent_statuses)):
                tdict = one.alyx.rest('tasks', 'partial_update', id=tdict['id'],
                                      data={'status': 'Held'})
            return tdict, registered_dsets
    # creates the job from the module name in the database
    exec_name = tdict['executable']
    strmodule, strclass = exec_name.rsplit('.', 1)
    classe = getattr(importlib.import_module(strmodule), strclass)
    task = classe(session_path, one=one, taskid=tdict['id'])
    # sets the status flag to started before running
    one.alyx.rest('tasks', 'partial_update', id=tdict['id'], data={'status': 'Started'})
    status = task.run()
    patch_data = {'time_elapsed_secs': task.time_elapsed_secs, 'log': task.log,
                  'version': task.version}
    # only registers successful runs
    if status == 0:
        # on a successful run, if there is no data to register, set status to Empty
        if task.outputs is None:
            patch_data['status'] = 'Empty'
        else:  # otherwise register data and set status to Complete
            try:
                registered_dsets = task.register_datasets(one=one, max_md5_size=max_md5_size)
            except BaseException:
                patch_data['status'] = 'Errored'
            patch_data['status'] = 'Complete'
    elif status == -1:
        patch_data['status'] = 'Errored'
    t = one.alyx.rest('tasks', 'partial_update', id=tdict['id'], data=patch_data)
    return t, registered_dsets



"""
Quality control on LFP
Based on code by Olivier Winter:
https://github.com/int-brain-lab/ibllib/blob/master/examples/ibllib/ephys_qc_raw.py

adapted by Anne Urai
Instructions https://docs.google.com/document/d/1lBNcssodWdBILBN0PrPWi0te4f6I8H_Bb-7ESxQdv5U/edit#
"""

from pathlib import Path

import sys
import glob
import os
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns


from ibllib.ephys import ephysqc
import alf.io
# from IPython import embed as shell


def _plot_spectra(outpath, typ, savefig=True):
    '''
    TODO document this function
    '''

    spec = alf.io.load_object(outpath, 'ephysQcFreq' + typ.upper(), namespace='spikeglx')

    # hack to ensure a single key name
    if 'power.probe_00' in spec.keys():
        spec['power'] = spec.pop('power.probe_00')
        spec['freq'] = spec.pop('freq.probe_00')
    elif 'power.probe_01' in spec.keys():
        spec['power'] = spec.pop('power.probe_01')
        spec['freq'] = spec.pop('freq.probe_01')

    # plot
    sns.set_style("whitegrid")
    plt.figure(figsize=[9, 4.5])
    ax = plt.axes()
    ax.plot(spec['freq'], 20 * np.log10(spec['power'] + 1e-14),
            linewidth=0.5, color=[0.5, 0.5, 0.5])
    ax.plot(spec['freq'], 20 * np.log10(np.median(spec['power'] + 1e-14, axis=1)), label='median')
    ax.set_xlabel(r'Frequency (Hz)')
    ax.set_ylabel(r'dB rel to $V^2.$Hz$^{-1}$')
    if typ == 'ap':
        ax.set_ylim([-275, -125])
    elif typ == 'lf':
        ax.set_ylim([-260, -60])
    ax.legend()
    ax.set_title(outpath)
    if savefig:
        plt.savefig(outpath / (typ + '_spec.png'), dpi=150)
        print('saved figure to %s' % (outpath / (typ + '_spec.png')))


def _plot_rmsmap(outpath, typ, savefig=True):
    '''
    TODO document this function
    '''

    rmsmap = alf.io.load_object(outpath, 'ephysQcTime' + typ.upper(), namespace='spikeglx')

    # hack to ensure a single key name
    if 'times.probe_00' in rmsmap.keys():
        rmsmap['times'] = rmsmap.pop('times.probe_00')
        rmsmap['rms'] = rmsmap.pop('rms.probe_00')
    elif 'times.probe_01' in rmsmap.keys():
        rmsmap['times'] = rmsmap.pop('times.probe_01')
        rmsmap['rms'] = rmsmap.pop('rms.probe_01')

    plt.figure(figsize=[12, 4.5])
    axim = plt.axes([0.2, 0.1, 0.7, 0.8])
    axrms = plt.axes([0.05, 0.1, 0.15, 0.8])
    axcb = plt.axes([0.92, 0.1, 0.02, 0.8])

    axrms.plot(np.median(rmsmap['rms'], axis=0)[:-1] * 1e6, np.arange(1, rmsmap['rms'].shape[1]))
    axrms.set_ylim(0, rmsmap['rms'].shape[1])

    im = axim.imshow(20 * np.log10(rmsmap['rms'].T + 1e-15), aspect='auto', origin='lower',
                     extent=[rmsmap['times'][0], rmsmap['times'][-1], 0, rmsmap['rms'].shape[1]])
    axim.set_xlabel(r'Time (s)')
    axrms.set_ylabel(r'Channel Number')
    plt.colorbar(im, cax=axcb)
    if typ == 'ap':
        im.set_clim(-110, -90)
        axrms.set_xlim(100, 0)
    elif typ == 'lf':
        im.set_clim(-100, -60)
        axrms.set_xlim(500, 0)
    axim.set_xlim(0, 4000)
    axim.set_title(outpath)
    if savefig:
        plt.savefig(outpath / (typ + '_rms.png'), dpi=150)


# ============================== ###
# FIND THE RIGHT FILES, RUN AS SCRIPT
# ============================== ###

if __name__ == '__main__':
    if len(sys.argv) != 2:
        print("Please give the folder path as an input argument!")
    else:
        outpath = Path(sys.argv[1])  # grab from command line input
        fbin = glob.glob(os.path.join(outpath, '*.lf.bin'))
        assert(len(fbin) > 0)
        print('fbin: %s' % fbin)
        # make sure you send a path for the time being and not a string
        ephysqc.extract_rmsmap(Path(fbin[0]))
        _plot_spectra(outpath, 'lf')
        _plot_rmsmap(outpath, 'lf')



import logging

from oneibl.one import ONE

log = logging.getLogger("ibllib")


def download_taskqc_raw_data(eid, one=None, fpga=False):
    """Download raw data required for performing task QC

    :param eid: A session UUID string
    :param one: An instance of ONE with which to download the data
    :param fpga: When True, downloads the raw ephys data required for extracting the FPGA task data
    :return: A list of file paths for the downloaded raw data
    """
    one = one or ONE()
    # Datasets required for extracting task data
    dstypes = [
        "_iblrig_taskData.raw",
        "_iblrig_taskSettings.raw",
        "_iblrig_encoderPositions.raw",
        "_iblrig_encoderEvents.raw",
        "_iblrig_stimPositionScreen.raw",
        "_iblrig_syncSquareUpdate.raw",
        "_iblrig_encoderTrialInfo.raw",
        "_iblrig_ambientSensorData.raw",
    ]
    # Extra files required for extracting task data from FPGA
    if fpga:
        dstypes.extend(['_spikeglx_sync.channels',
                        '_spikeglx_sync.polarities',
                        '_spikeglx_sync.times',
                        'ephysData.raw.meta',
                        'ephysData.raw.wiring'])
    # Download the data via ONE
    return one.load(eid, dataset_types=dstypes, download_only=True)
